#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import subprocess
import unittest2

from neutron.tests import base

from gbpservice.contrib.nfp.configurator.agents import generic_config as gc
from gbpservice.contrib.nfp.configurator.lib import (
                                    generic_config_constants as const)
from gbpservice.contrib.nfp.configurator.lib import constants as common_const
from gbpservice.contrib.tests.unit.nfp.configurator.test_data import (
                                                        fw_test_data as fo)


class GenericConfigRpcManagerTestCase(base.BaseTestCase):
    """ Implement test cases for RPC manager methods of generic config agent.

    """

    def __init__(self, *args, **kwargs):
        super(GenericConfigRpcManagerTestCase, self).__init__(
                                                        *args, **kwargs)
        self.fo = fo.FakeObjects()

    @mock.patch(__name__ + '.fo.FakeObjects.sc')
    @mock.patch(__name__ + '.fo.FakeObjects.conf')
    def _get_GenericConfigRpcManager_object(self, conf, sc):
        """ Retrieves RPC manager object of generic config agent.

        :param sc: mocked service controller object of process model framework
        :param conf: mocked OSLO configuration file

        Returns: object of generic config's RPC manager
        and service controller.

        """

        agent = gc.GenericConfigRpcManager(sc, conf)
        return agent, sc

    def _test_event_creation(self, method):
        """ Tests event creation and enqueueing for create/delete
        operation of generic config agent's RPC manager.

        :param method: CONFIGURE_INTERFACES/CLEAR_INTERFACES/
        CONFIGURE_ROUTES/CLEAR_ROUTES

        Returns: none

        """

        agent, sc = self._get_GenericConfigRpcManager_object()
        arg_dict = {'context': self.fo.context,
                    'resource_data': self.fo._fake_resource_data()}
        with mock.patch.object(
                    sc, 'new_event', return_value='foo') as mock_sc_event, (
             mock.patch.object(sc, 'post_event')) as mock_sc_rpc_event:
            call_method = getattr(agent, method.lower())

            call_method(self.fo.context, self.fo._fake_resource_data())

            if 'HEALTHMONITOR' in method:
                mock_sc_event.assert_called_with(id=method,
                                                 data=arg_dict,
                                                 key=self.fo.vmid)
            else:
                mock_sc_event.assert_called_with(id=method,
                                                 data=arg_dict, key=None)
            mock_sc_rpc_event.assert_called_with('foo')

    def test_configure_interfaces_genericconfigrpcmanager(self):
        """ Implements test case for configure interfaces method
        of generic config agent RPCmanager.

        Returns: none

        """

        self._test_event_creation(const.EVENT_CONFIGURE_INTERFACES)

    def test_clear_interfaces_genericconfigrpcmanager(self):
        """ Implements test case for clear interfaces method
        of generic config agent RPCmanager.

        Returns: none

        """

        self._test_event_creation(const.EVENT_CLEAR_INTERFACES)

    def test_configure_routes_genericconfigrpcmanager(self):
        """ Implements test case for configure routes method
        of generic config agent RPCmanager.

        Returns: none

        """

        self._test_event_creation(const.EVENT_CONFIGURE_ROUTES)

    def test_clear_routes_genericconfigrpcmanager(self):
        """ Implements test case for clear routes method
        of generic config agent RPCmanager.

        Returns: none

        """

        self._test_event_creation(const.EVENT_CLEAR_ROUTES)

    def test_configure_hm_genericconfigrpcmanager(self):
        """ Implements test case for configure healthmonitor method
        of generic config agent RPCmanager.

        Returns: none

        """

        self._test_event_creation(const.EVENT_CONFIGURE_HEALTHMONITOR)

    @unittest2.skip('not implemented yet')
    def test_clear_hm_genericconfigrpcmanager(self):
        """ Implements test case for clear healthmonitor method
        of generic config agent RPCmanager.

        Returns: none

        """

        self._test_event_creation(const.EVENT_CLEAR_HEALTHMONITOR)


class GenericConfigEventHandlerTestCase(base.BaseTestCase):
    """ Implements test cases for event handler methods
    of generic config agent.

    """

    def __init__(self, *args, **kwargs):
        super(GenericConfigEventHandlerTestCase, self).__init__(
                                                        *args, **kwargs)
        self.fo = fo.FakeObjects()
        self.empty = self.fo.empty_dict
        self.context = {'notification_data': {},
                        'resource': 'interfaces'}

    @mock.patch(__name__ + '.fo.FakeObjects.rpcmgr')
    @mock.patch(__name__ + '.fo.FakeObjects.drivers')
    @mock.patch(__name__ + '.fo.FakeObjects.sc')
    def _get_GenericConfigEventHandler_object(self, sc, drivers, rpcmgr):
        """ Retrieves event handler object of generic config.

        :param sc: mocked service controller object of process model framework
        :param rpcmgr: object of configurator's RPC manager
        :param drivers: list of driver objects for firewall agent

        Returns: object of generic config's event handler

        """

        agent = gc.GenericConfigEventHandler(sc, drivers, rpcmgr)
        return agent, sc

    def _test_handle_event(self, ev):
        """ Test handle event method of generic config agent for various
        device configuration operations.

        :param ev: fake event data which has to be actually sent by
        process framework.

        Returns: None

        """

        agent, sc = self._get_GenericConfigEventHandler_object()
        driver = mock.Mock()

        with mock.patch.object(
                driver, const.EVENT_CONFIGURE_INTERFACES.lower(),
                return_value=common_const.SUCCESS) as (mock_config_inte), (
            mock.patch.object(
                driver, const.EVENT_CLEAR_INTERFACES.lower(),
                return_value=common_const.SUCCESS)) as (mock_clear_inte), (
            mock.patch.object(
                driver, const.EVENT_CONFIGURE_ROUTES.lower(),
                return_value=common_const.SUCCESS)) as (
                                                    mock_config_src_routes), (
            mock.patch.object(
                driver, const.EVENT_CLEAR_ROUTES.lower(),
                return_value=common_const.SUCCESS)) as (
                                                mock_delete_src_routes), (
            mock.patch.object(
                sc, 'poll_event')) as mock_hm_poll_event, (
            mock.patch.object(
                driver, const.EVENT_CONFIGURE_HEALTHMONITOR.lower(),
                return_value=common_const.SUCCESS)), (
            mock.patch.object(
                driver, const.EVENT_CLEAR_HEALTHMONITOR.lower(),
                return_value=common_const.SUCCESS)) as mock_clear_hm, (
            mock.patch.object(
                agent, '_get_driver', return_value=driver)):

            if const.EVENT_CONFIGURE_HEALTHMONITOR in ev.id:
                ev.id, periodicity = ev.id.split()

            agent.handle_event(ev)

            resource_data = self.fo._fake_resource_data()
            if ev.id == const.EVENT_CONFIGURE_INTERFACES:
                mock_config_inte.assert_called_with(
                                        self.fo.context, resource_data)
            elif ev.id == const.EVENT_CLEAR_INTERFACES:
                mock_clear_inte.assert_called_with(
                                        self.fo.context, resource_data)
            elif ev.id == const.EVENT_CONFIGURE_ROUTES:
                mock_config_src_routes.assert_called_with(
                            self.fo.context, resource_data)
            elif ev.id == const.EVENT_CLEAR_ROUTES:
                mock_delete_src_routes.assert_called_with(
                            self.fo.context, resource_data)
            elif const.EVENT_CONFIGURE_HEALTHMONITOR in ev.id:
                if periodicity == const.INITIAL:
                    mock_hm_poll_event.assert_called_with(
                        ev,
                        max_times=const.EVENT_CONFIGURE_HEALTHMONITOR_MAXRETRY)
                elif periodicity == const.FOREVER:
                    mock_hm_poll_event.assert_called_with(ev)
            elif ev.id == const.EVENT_CLEAR_HEALTHMONITOR:
                mock_clear_hm.assert_called_with(
                            self.fo.context, resource_data)

    def _test_handle_periodic_event(self, ev):
        """ Test handle periodic event method of generic config agent
        for healthmonitor configuration.

        :param ev: fake event data which has to be actually sent by
        process framewrok.

        Returns: None

        """

        agent, _ = self._get_GenericConfigEventHandler_object()
        driver = mock.Mock()

        with mock.patch.object(
                agent, '_get_driver', return_value=driver), (
             mock.patch.object(
                    driver, const.EVENT_CONFIGURE_HEALTHMONITOR.lower(),
                    return_value=common_const.SUCCESS)) as mock_dvr, (
             mock.patch.object(subprocess,
                               'check_output', return_value=True)):

            agent.handle_configure_healthmonitor(ev)

            self.assertEqual(mock_dvr.return_value, common_const.SUCCESS)

    def test_configure_interfaces_genericconfigeventhandler(self):
        """ Implements test case for configure interfaces method
        of generic config event handler.

        Returns: none

        """

        ev = fo.FakeEventGenericConfig()
        ev.id = const.EVENT_CONFIGURE_INTERFACES
        self._test_handle_event(ev)

    def test_clear_interfaces_genericconfigeventhandler(self):
        """ Implements test case for clear interfaces method
        of generic config event handler.

        Returns: none

        """

        ev = fo.FakeEventGenericConfig()
        ev.id = const.EVENT_CLEAR_INTERFACES
        self._test_handle_event(ev)

    def test_configure_routes_genericconfigeventhandler(self):
        """ Implements test case for configure routes method
        of generic config event handler.

        Returns: none

        """

        ev = fo.FakeEventGenericConfig()
        ev.id = const.EVENT_CONFIGURE_ROUTES
        self._test_handle_event(ev)

    def test_clear_routes_genericconfigeventhandler(self):
        """ Implements test case for clear routes method
        of generic config event handler.

        Returns: none

        """

        ev = fo.FakeEventGenericConfig()
        ev.id = const.EVENT_CLEAR_ROUTES
        self._test_handle_event(ev)

    def test_configure_hm_initial_genericconfigeventhandler(self):
        """ Implements test case for configure health monitor method
         with specified polling in generic config event handler.

        Returns: none

        """

        ev = fo.FakeEventGenericConfig()
        ev.id = 'CONFIGURE_HEALTHMONITOR initial'
        self._test_handle_event(ev)

    def test_configure_hm_forever_genericconfigeventhandler(self):
        """ Implements test case for configure health monitor method
        with forever polling in generic config event handler.

        Returns: none

        """

        ev = fo.FakeEventGenericConfig()
        ev.data['resource_data']['nfds'][0].update(
                                            {'periodicity': const.FOREVER})
        ev.id = 'CONFIGURE_HEALTHMONITOR forever'
        self._test_handle_event(ev)

    def test_clear_hm_genericconfigeventhandler(self):
        """ Implements test case for clear health monitor method
        of generic config event handler.

        Returns: none
        """

        ev = fo.FakeEventGenericConfig()
        ev.id = 'CLEAR_HEALTHMONITOR'
        self._test_handle_event(ev)

    def test_handle_configure_healthmonitor_genericconfigeventhandler(self):
        """ Implements test case for handle configure health monitor
         method of generic config event handler.

        Returns: none

        """

        ev = fo.FakeEventGenericConfig()
        ev.id = const.EVENT_CONFIGURE_HEALTHMONITOR
        self._test_handle_periodic_event(ev)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock

from gbpservice.contrib.nfp.configurator.agents import vpn
from gbpservice.contrib.nfp.configurator.lib import vpn_constants as const
from gbpservice.contrib.tests.unit.nfp.configurator.test_data import (
    vpn_test_data)

from neutron.tests import base


class VPNaasRpcManagerTestCase(base.BaseTestCase):
    """
    Implements test cases for RPC manager methods of vpn agent
    """
    def __init__(self, *args, **kwargs):
        super(VPNaasRpcManagerTestCase, self).__init__(*args, **kwargs)
        self.test_dict = vpn_test_data.VPNTestData()
        self.conf = self.test_dict.conf
        self.sc = mock.Mock()
        self.rpcmgr = vpn.VPNaasRpcManager(self.conf, self.sc)

    def test_vpnservice_updated(self):
        resource_data = self.test_dict._create_ipsec_site_conn_obj()
        with mock.patch.object(self.sc, 'new_event',
                               return_value='foo'), (
             mock.patch.object(self.sc, 'post_event')) as mock_post_event:
            self.rpcmgr.vpnservice_updated(
                        self.test_dict.make_service_context(),
                        resource_data=resource_data)
            mock_post_event.assert_called_with('foo')


class VPNaasEventHandlerTestCase(base.BaseTestCase):
    """
    Implements test cases for RPC manager methods of vpn agent
    """
    def __init__(self, *args, **kwargs):
        super(VPNaasEventHandlerTestCase, self).__init__(*args, **kwargs)
        self.test_dict = vpn_test_data.VPNTestData()
        self.sc = self.test_dict.sc
        self.conf = self.test_dict.conf
        self.handler = vpn.VPNaasEventHandler(self.test_dict.sc,
                                              self.test_dict.drivers)
        self.ev = vpn_test_data.FakeEvent()
        self.rpc_sender = vpn.VpnaasRpcSender(self.sc)
        self.driver = mock.Mock()

    def test_handle_event(self):
        """
        Test to handle the vpn agent's vpnservice_updated method to
        handle various vpn operations

        """
        with mock.patch.object(self.handler,
                               '_get_driver',
                               return_value=self.test_dict.drivers), (
             mock.patch.object(
                     self.driver,
                     'vpnservice_updated')) as mock_vpnservice_updated:
            self.handler._vpnservice_updated(self.ev, self.driver)
            mock_vpnservice_updated.assert_called_with(self.ev.data['context'],
                                                       self.ev.data[
                                                           'resource_data'])

    def test_sync(self):
        """
        Test to handle the vpn service status like ACTIVE, ERROR
        after the configurations.

        """
        with mock.patch.object(self.rpc_sender,
                               'get_vpn_servicecontext'), (
             mock.patch.object(self.driver,
                               'check_status',
                               return_value=const.STATE_ACTIVE)):

            self.handler.service_driver = self.driver
            self.assertEqual(self.handler.sync(self.ev), {'poll': False})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.api import extensions
from neutron_lib import constants

from gbpservice.neutron.extensions import group_policy as gp


EXTENDED_ATTRIBUTES_2_0 = {
    gp.POLICY_TARGETS: {
        'pt_extension': {'allow_post': True,
                         'allow_put': True,
                         'default': constants.ATTR_NOT_SPECIFIED,
                         'is_visible': True,
                         'enforce_policy': True},
    },
    gp.POLICY_TARGET_GROUPS: {
        'ptg_extension': {'allow_post': True,
                          'allow_put': True,
                          'default': constants.ATTR_NOT_SPECIFIED,
                          'is_visible': True,
                          'enforce_policy': True},
    },
    gp.L2_POLICIES: {
        'l2p_extension': {'allow_post': True,
                          'allow_put': True,
                          'default': constants.ATTR_NOT_SPECIFIED,
                          'is_visible': True,
                          'enforce_policy': True},
    },
    gp.L3_POLICIES: {
        'l3p_extension': {'allow_post': True,
                          'allow_put': True,
                          'default': constants.ATTR_NOT_SPECIFIED,
                          'is_visible': True,
                          'enforce_policy': True},
    },
    gp.POLICY_CLASSIFIERS: {
        'pc_extension': {'allow_post': True,
                         'allow_put': True,
                         'default': constants.ATTR_NOT_SPECIFIED,
                         'is_visible': True,
                         'enforce_policy': True},
    },
    gp.POLICY_ACTIONS: {
        'pa_extension': {'allow_post': True,
                         'allow_put': True,
                         'default': constants.ATTR_NOT_SPECIFIED,
                         'is_visible': True,
                         'enforce_policy': True},
    },
    gp.POLICY_RULES: {
        'pr_extension': {'allow_post': True,
                         'allow_put': True,
                         'default': constants.ATTR_NOT_SPECIFIED,
                         'is_visible': True,
                         'enforce_policy': True},
    },
    gp.POLICY_RULE_SETS: {
        'prs_extension': {'allow_post': True,
                          'allow_put': True,
                          'default': constants.ATTR_NOT_SPECIFIED,
                          'is_visible': True,
                          'enforce_policy': True},
    },
    gp.NETWORK_SERVICE_POLICIES: {
        'nsp_extension': {'allow_post': True,
                          'allow_put': True,
                          'default': constants.ATTR_NOT_SPECIFIED,
                          'is_visible': True,
                          'enforce_policy': True},
    },
    gp.EXTERNAL_SEGMENTS: {
        'es_extension': {'allow_post': True,
                         'allow_put': True,
                         'default': constants.ATTR_NOT_SPECIFIED,
                         'is_visible': True,
                         'enforce_policy': True},
    },
    gp.EXTERNAL_POLICIES: {
        'ep_extension': {'allow_post': True,
                         'allow_put': True,
                         'default': constants.ATTR_NOT_SPECIFIED,
                         'is_visible': True,
                         'enforce_policy': True},
    },
    gp.NAT_POOLS: {
        'np_extension': {'allow_post': True,
                         'allow_put': True,
                         'default': constants.ATTR_NOT_SPECIFIED,
                         'is_visible': True,
                         'enforce_policy': True},
    },
}


class Test_extension(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "group policy test extension"

    @classmethod
    def get_alias(cls):
        return "test_extension"

    @classmethod
    def get_description(cls):
        return _("Adds test attributes to group policy resources.")

    @classmethod
    def get_namespace(cls):
        return ("http://docs.openstack.org/ext/neutron/grouppolicy/test/"
                "test_extension/api/v1.0")

    @classmethod
    def get_updated(cls):
        return "2014-10-24T10:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from gbpservice._i18n import _LE
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import exceptions

LOG = log.getLogger(__name__)

PROJECT_TYPE_TAG = 'prj'
NETWORK_TYPE_TAG = 'net'
ADDRESS_SCOPE_TYPE_TAG = 'as'
ROUTER_TYPE_TAG = 'rtr'
L3_POLICY_TYPE_TAG = 'l3p'
POLICY_RULE_SET_TYPE_TAG = 'prs'
POLICY_RULE_TYPE_TAG = 'pr'
APPLICATION_POLICY_GROUP_TYPE_TAG = 'apg'
PORT_PAIR_GROUP_TYPE_TAG = 'ppg'
PORT_PAIR_TYPE_TAG = 'pp'
PORT_TYPE_TAG = 'prt'
FLOW_CLASSIFIER_TYPE_TAG = 'flc'
PORT_CHAIN_TYPE_TAG = 'ptc'


class APICNameMapper(object):
    # This class may be overriden to customize the mapping from
    # OpenStack resource IDs to APIC resource names. Prefixes can be
    # supplied in order to allow mapping a single OpenStack resource
    # to multiple APIC resources of the same type. There are reverse
    # mapping methods for most resources, but there is no
    # reverse_project method because of potential use cases for
    # mapping multiple OpenStack project IDs to the same APIC Tenant.

    def project(self, session, id, prefix=""):
        # REVISIT: The external connectiviy unit tests pass "common"
        # as a project_id, and expect this to be mapped to the AIM
        # common Tenant. Its not clear how a real deployment using
        # Keystone would arrange to have "common" as a valid
        # project_id, so maybe this special handling is not really
        # needed.
        return (id if id == "common" else
                self._map(session, id, PROJECT_TYPE_TAG, prefix))

    def network(self, session, id, prefix=""):
        return self._map(session, id, NETWORK_TYPE_TAG, prefix)

    def reverse_network(self, session, name, prefix="", enforce=True):
        return self._unmap(
            session, name, NETWORK_TYPE_TAG, prefix, enforce)

    def address_scope(self, session, id, prefix=""):
        return self._map(session, id, ADDRESS_SCOPE_TYPE_TAG, prefix)

    def reverse_address_scope(self, session, name, prefix="", enforce=True):
        return self._unmap(
            session, name, ADDRESS_SCOPE_TYPE_TAG, prefix, enforce)

    def router(self, session, id, prefix=""):
        return self._map(session, id, ROUTER_TYPE_TAG, prefix)

    def reverse_router(self, session, name, prefix="", enforce=True):
        return self._unmap(session, name, ROUTER_TYPE_TAG, prefix, enforce)

    def l3_policy(self, session, id, prefix=""):
        return self._map(session, id, L3_POLICY_TYPE_TAG, prefix)

    def reverse_l3_policy(self, session, name, prefix="", enforce=True):
        return self._unmap(
            session, name, L3_POLICY_TYPE_TAG, prefix, enforce)

    def policy_rule_set(self, session, id, prefix=""):
        return self._map(session, id, POLICY_RULE_SET_TYPE_TAG, prefix)

    def reverse_policy_rule_set(self, session, name, prefix="", enforce=True):
        return self._unmap(
            session, name, POLICY_RULE_SET_TYPE_TAG, prefix, enforce)

    def policy_rule(self, session, id, prefix=""):
        return self._map(session, id, POLICY_RULE_TYPE_TAG, prefix)

    def reverse_policy_rule(self, session, name, prefix="", enforce=True):
        return self._unmap(
            session, name, POLICY_RULE_TYPE_TAG, prefix, enforce)

    def application_policy_group(self, session, id, prefix=""):
        return self._map(
            session, id, APPLICATION_POLICY_GROUP_TYPE_TAG, prefix)

    def reverse_application_policy_group(
        self, session, name, prefix="", enforce=True):
        return self._unmap(
            session, name, APPLICATION_POLICY_GROUP_TYPE_TAG, prefix, enforce)

    def port_pair_group(self, session, id, prefix=""):
        return self._map(session, id, PORT_PAIR_GROUP_TYPE_TAG, prefix)

    def reverse_port_pair_group(self, session, name, prefix="", enforce=True):
        return self._unmap(session, name, PORT_PAIR_GROUP_TYPE_TAG, prefix,
                           enforce)

    def port_pair(self, session, id, prefix=""):
        return self._map(session, id, PORT_PAIR_TYPE_TAG, prefix)

    def reverse_port_pair(self, session, name, prefix="", enforce=True):
        return self._unmap(session, name, PORT_PAIR_TYPE_TAG, prefix, enforce)

    def port(self, session, id, prefix=""):
        return self._map(session, id, PORT_TYPE_TAG, prefix)

    def reverse_port(self, session, name, prefix="", enforce=True):
        return self._unmap(session, name, PORT_TYPE_TAG, prefix, enforce)

    def flow_classifier(self, session, id, prefix=""):
        return self._map(session, id, FLOW_CLASSIFIER_TYPE_TAG, prefix)

    def reverse_flow_classifier(self, session, name, prefix="", enforce=True):
        return self._unmap(session, name, FLOW_CLASSIFIER_TYPE_TAG, prefix,
                           enforce)

    def port_chain(self, session, id, prefix=""):
        return self._map(session, id, PORT_CHAIN_TYPE_TAG, prefix)

    def reverse_port_chain(self, session, name, prefix="", enforce=True):
        return self._unmap(session, name, PORT_CHAIN_TYPE_TAG, prefix, enforce)

    def _map(self, session, id, type_tag, prefix):
        return ("%(prefix)s%(type_tag)s_%(id)s" %
                {'prefix': prefix, 'type_tag': type_tag, 'id': id})

    def _unmap(self, session, name, type_tag, prefix, enforce):
        pos = len(prefix) + len(type_tag) + 1
        if self._map(session, "", type_tag, prefix) == name[:pos]:
            return name[pos:]
        elif enforce:
            msg = _LE("Attempted to reverse-map invalid APIC name '%s'") % name
            raise exceptions.InternalError(details=msg)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""svi-apic

Revision ID: 5fe2285071fd
Revises: 886c376885a3
Create Date: 2018-01-05 00:00:00.000000

"""

# revision identifiers, used by Alembic.
revision = '5fe2285071fd'
down_revision = '886c376885a3'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('apic_aim_network_extensions',
                  sa.Column('svi', sa.Boolean))
    op.add_column('apic_aim_network_mappings',
                  sa.Column('l3out_name', sa.String(64), nullable=True))
    op.add_column('apic_aim_network_mappings',
                  sa.Column('l3out_ext_net_name', sa.String(64),
                            nullable=True))
    op.add_column('apic_aim_network_mappings',
                  sa.Column('l3out_tenant_name', sa.String(64),
                            nullable=True))


def downgrade():
    pass



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron_lib.db import model_base
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import orm

from gbpservice.nfp.common import constants as nfp_constants

TENANT_ID_MAX_LEN = 255
DESCRIPTION_MAX_LEN = 4096


class HasStatus(object):
    """status mixin, add to subclasses that have a status."""

    status = sa.Column(sa.String(16), nullable=False, index=True)


class HasStatusDescription(HasStatus):
    """Status with description mixin."""

    status_description = sa.Column(sa.String(DESCRIPTION_MAX_LEN))


BASE = declarative_base(cls=model_base.NeutronBaseV2)


class PortInfo(BASE, model_base.HasId, model_base.HasProject):
    """Represents the Port Information"""
    __tablename__ = 'nfp_port_infos'

    port_model = sa.Column(sa.Enum(nfp_constants.NEUTRON_PORT,
                                   nfp_constants.GBP_PORT,
                                   name='port_model'))
    port_classification = sa.Column(sa.Enum(nfp_constants.PROVIDER,
                                            nfp_constants.CONSUMER,
                                            nfp_constants.MANAGEMENT,
                                            nfp_constants.MONITOR,
                                            name='port_classification'))
    port_role = sa.Column(sa.Enum(nfp_constants.ACTIVE_PORT,
                                  nfp_constants.STANDBY_PORT,
                                  nfp_constants.MASTER_PORT,
                                  name='port_role'),
                          nullable=True)


class NetworkInfo(BASE, model_base.HasId, model_base.HasProject):
    """Represents the Network Service Instance"""
    __tablename__ = 'nfp_network_infos'

    network_model = sa.Column(sa.Enum(nfp_constants.NEUTRON_NETWORK,
                                      nfp_constants.GBP_NETWORK,
                                      name='network_model'),
                              nullable=False)


class NSIPortAssociation(BASE):
    """One to many relation between NSIs and DataPorts."""
    __tablename__ = 'nfp_nfi_dataport_associations'

    network_function_instance_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('nfp_network_function_instances.id'), primary_key=True)
    data_port_id = sa.Column(sa.String(36),
                             sa.ForeignKey('nfp_port_infos.id',
                                           ondelete='CASCADE'),
                             primary_key=True)


class NetworkFunctionInstance(BASE, model_base.HasId, model_base.HasProject,
                              HasStatusDescription):
    """Represents the Network Function Instance"""
    __tablename__ = 'nfp_network_function_instances'

    name = sa.Column(sa.String(255))
    description = sa.Column(sa.String(255))
    ha_state = sa.Column(sa.String(255))
    network_function_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('nfp_network_functions.id', ondelete="SET NULL"),
        nullable=True)
    network_function_device_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('nfp_network_function_devices.id', ondelete="SET NULL"),
        nullable=True)
    port_info = orm.relationship(
        NSIPortAssociation,
        cascade='all, delete-orphan')


class NetworkFunction(BASE, model_base.HasId, model_base.HasProject,
                      HasStatusDescription):
    """Represents the Network Function object"""
    __tablename__ = 'nfp_network_functions'

    name = sa.Column(sa.String(255))
    description = sa.Column(sa.String(1024))
    service_id = sa.Column(sa.String(36), nullable=False)
    service_chain_id = sa.Column(sa.String(36), nullable=True)
    service_profile_id = sa.Column(sa.String(36), nullable=False)
    service_config = sa.Column(sa.TEXT)
    config_policy_id = sa.Column(sa.String(36), nullable=True)
    network_function_instances = orm.relationship(
        NetworkFunctionInstance,
        backref='network_function')


class NetworkFunctionDevice(BASE, model_base.HasId, model_base.HasProject,
                            HasStatusDescription):
    """Represents the Network Function Device"""
    __tablename__ = 'nfp_network_function_devices'

    name = sa.Column(sa.String(255))
    description = sa.Column(sa.String(255))
    mgmt_ip_address = sa.Column(sa.String(36), nullable=True)
    mgmt_port_id = sa.Column(sa.String(36),
                             sa.ForeignKey('nfp_port_infos.id',
                                           ondelete='SET NULL'),
                             nullable=True)
    monitoring_port_id = sa.Column(sa.String(36),
                                   sa.ForeignKey('nfp_port_infos.id',
                                                 ondelete='SET NULL'),
                                   nullable=True)
    monitoring_port_network = sa.Column(sa.String(36),
                                        sa.ForeignKey('nfp_network_infos.id',
                                                      ondelete='SET NULL'),
                                        nullable=True)
    service_vendor = sa.Column(sa.String(36), nullable=False, index=True)
    max_interfaces = sa.Column(sa.Integer(), nullable=False)
    reference_count = sa.Column(sa.Integer(), nullable=False)
    interfaces_in_use = sa.Column(sa.Integer(), nullable=False)
    provider_metadata = sa.Column(sa.String(1024), nullable=True)
    gateway_port = sa.Column(sa.String(36), nullable=True)


class ClusterInfo(BASE, model_base.HasId, model_base.HasProject):
    """
    This table contains info about the ports participating in
    cluster and optional.
    """
    __tablename__ = 'nfd_cluster_mapping_info'
    network_function_device_id = sa.Column(sa.String(36), nullable=False)
    cluster_group = sa.Column(sa.Integer(), nullable=True)
    virtual_ip = sa.Column(sa.String(36), nullable=True)
    multicast_ip = sa.Column(sa.String(36), nullable=True)
    cluster_name = sa.Column(sa.String(36), nullable=True)


class ServiceGatewayDetails(BASE, model_base.HasId):
    __tablename__ = 'nfp_service_gateway_info'
    network_function_id = sa.Column(sa.String(36), sa.ForeignKey(
        'nfp_network_functions.id', ondelete='CASCADE'), nullable=False,
        primary_key=True)
    gateway_ptg = sa.Column(sa.String(36), nullable=False)
    primary_instance_gw_pt = sa.Column(sa.String(36), nullable=True)
    secondary_instance_gw_pt = sa.Column(sa.String(36), nullable=True)
    primary_gw_vip_pt = sa.Column(sa.String(36), nullable=True)
    secondary_gw_vip_pt = sa.Column(sa.String(36), nullable=True)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import os
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():

    # Run the Testcase:
    test = test_gbp_pc_func()
    if test.test_gbp_pc_func_1() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PC_FUNC_1')
    if test.test_gbp_pc_func_2() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PC_FUNC_2')
    if test.test_gbp_pc_func_3() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PC_FUNC_3')
    if test.test_gbp_pc_func_4() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PC_FUNC_4')
    test.cleanup()
    utils_libs.report_results('test_gbp_pc_func', 'test_results.txt')
    sys.exit(1)


class test_gbp_pc_func(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    cmd = 'rm /tmp/test_gbp_pc_func.log'
    commands.getoutput(cmd)
    hdlr = logging.FileHandler('/tmp/test_gbp_pc_func.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self):
        """
        Init def
        """
        self._log.info(
            "\n## START OF GBP POLICY_CLASSIFIER FUNCTIONALITY TESTSUITE\n")
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.cls_name = 'demo_pc'
        self.act_name = 'demo_pa'

    def cleanup(self, tc_name=''):
        if tc_name != '':
            self._log.info('%s: FAILED' % (tc_name))
        for obj in ['rule', 'classifier', 'action']:
            self.gbpcfg.gbp_del_all_anyobj(obj)

    def test_gbp_pc_func_1(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_PC_FUNC_1: CREATE/VERIFY/DELETE/VERIFY a POLICY "
            "CLASSIFIER with DEFAULT ATTRIB VALUE\n"
            "TEST_STEPS::\n"
            "Create Policy Classifier Object,default params\n"
            "Verify the attributes & value, show & list cmds\n"
            "Delete Policy Classifier using Name\n"
            "Verify that PC has got deleted, show & list cmds\n"
            "Recreate Policy Classifier Object inorder to test Delete "
            "using UUID\n"
            "Delete using UUID\n"
            "Verify that PC has got deleted, show & list cmds\n"
            "##############################################################\n")
        # Testcase work-flow starts
        self._log.info(
            "\n## Step 1: Create Classifier with default attrib vals##\n")
        cls_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'classifier', self.cls_name)
        if cls_uuid != 0:
            self._log.info(
                "\n## Step 1: Create Classifier Passed, UUID == %s\n" %
                (cls_uuid))
        else:
            self._log.info("\n## Step 1: Create Classifier == Failed")
            return 0
        self._log.info("\n## Step 4A: Verify Classifier using -list cmd")
        if self.gbpverify.gbp_classif_verify(0, self.cls_name, cls_uuid) == 0:
            self._log.info(
                "\n## Step 4A: Verify Classifier using -list option == Failed")
            return 0
        self._log.info("\n## Step 4B: Verify Classifier using -show cmd")
        if self.gbpverify.gbp_classif_verify(
                1, self.cls_name, id=cls_uuid) == 0:
            self._log.info(
                "\n## Step 4B: Verify Classifier using -show option == Failed")
            return 0

        ######
        self._log.info("\n## Step 3: Delete Classifier using name  ##")
        if self.gbpcfg.gbp_policy_cfg_all(0, 'classifier', self.cls_name):
            self._log.info(
                "\n## Step 3: Delete Classifier using Name == %s, Passed" %
                (self.cls_name))
        else:
            self._log.info(
                "\n## Step 3: Delete Classifier using Name == Failed")
            return 0
        if self.gbpverify.gbp_classif_verify(0, self.cls_name, cls_uuid) != 0:
            self._log.info(
                "\n## Step 3A: Verify Classifier is Deleted using -list "
                "option == Failed")
            return 0
        if self.gbpverify.gbp_classif_verify(
                1, self.cls_name, id=cls_uuid, shared='False') != 0:
            self._log.info(
                "\n## Step 3B: Verify Classifier is Deleted using -show "
                "option == Failed")
            return 0

        cls_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'classifier', self.cls_name)
        if cls_uuid:
            self._log.info(
                "\n## Step 4: Re-created a Policy Classifier with default "
                "inorder to delete with ID")
            self._log.info("\n## Step 5: Delete Classifier using UUID  ##")
            if self.gbpcfg.gbp_policy_cfg_all(0, 'classifier', cls_uuid):
                self._log.info(
                    "\n## Step 5: Delete Classifier Passed using UUID == %s" %
                    (cls_uuid))
            else:
                self._log.info(
                    "\n##  Step 5: Delete Classifier using UUID == Failed")
            if self.gbpverify.gbp_classif_verify(
                    0, self.cls_name, cls_uuid) != 0:
                self._log.info(
                    "\n##  Step 5A: Verify Classifier is Deleted using "
                    "-list option == Failed")
                return 0
            if self.gbpverify.gbp_classif_verify(
                    1, self.cls_name, id=cls_uuid, shared='False') != 0:
                self._log.info(
                    "\n## Step 5B: Verify Classifier is Deleted using "
                    "-show option == Failed")
                return 0
            self._log.info(
                "\n## Step 5: Delete of Policy Classifier using "
                "UUID == Passed")
        else:
            self._log.info(
                "\n## Step 6: Recreate of Policy Classifier using "
                "Default == Failed")
            return 0
        self._log.info("\n## TESTCASE_GBP_PC_FUNC_1: PASSED")

    def test_gbp_pc_func_2(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_PC_FUNC_2: UPDATE/VERIFY/DELETE/VERIFY EACH "
            "ATTRIB of a POLICY CLASSIFIER\n"
            "TEST_STEP::\n"
            "Create Policy Classifier using Default param values\n"
            "Update Each the Polciy Classifier's editable params one "
            "at a time\n"
            "Verify the Policy Classifier's attributes & values, show "
            "& list cmds\n"
            "Delete the Policy Classifier\n"
            "Verify Policy Classifier successfully deleted\n"
            "##############################################################\n")
        # Testcase work-flow starts
        self._log.info(
            "\n## Step 1: Create Classifier with default attrib vals##\n")
        cls_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'classifier', self.cls_name)
        if cls_uuid != 0:
            self._log.info(
                "\n## Step 1: Create Classifier Passed, UUID == %s\n" %
                (cls_uuid))
        else:
            self._log.info("\n## Step 1: Create Classifier == Failed")
            return 0
        # for attr,val in attrib.iteritems():
        self._log.info(
            "\n## Step 2: Update Policy Classifier attributes one at a "
            "time %s ##")
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'classifier', cls_uuid, name='grppol_pc') == 0:
            self._log.info(
                "\n## Step 2: Updating Policy Classifier's attribute "
                "Name, Failed")
            return 0
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'classifier', cls_uuid, protocol='tcp') == 0:
            self._log.info(
                "\n## Step 2: Updating Policy Classifier's attribute "
                "Protocol, Failed")
            return 0
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'classifier', cls_uuid, direction='bi') == 0:
            self._log.info(
                "\n## Step 2: Updating Policy Classifier's attribute "
                "Direction, Failed")
            return 0
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'classifier', cls_uuid, port_range='22:1022') == 0:
            self._log.info(
                "\n## Step 2: Updating Policy Classifier's attribute "
                "Port Range, Failed")
            return 0
        if self.gbpverify.gbp_classif_verify(
                0, 'grppol_pc', cls_uuid, 'tcp', 'bi', '22:1022') == 0:
            self._log.info(
                "\n## Step 2A: Verify Policy Classifier Updated Attributes "
                "using -list option == Failed")
            return 0
        if self.gbpverify.gbp_classif_verify(
                1,
                'grppol_pc',
                id=cls_uuid,
                protocol='tcp',
                port_range='22:1022',
                direction='bi') == 0:
            self._log.info(
                "\n## Step 2B: Verify Policy Classifier Updated Attributes "
                "using -show option == Failed")
            return 0
        if self.gbpcfg.gbp_policy_cfg_all(0, 'classifier', cls_uuid):
            self._log.info(
                "\n## Step 3: Deleted the Classifier == %s\n" %
                (cls_uuid))
        else:
            self._log.info(
                "\n## Step 3: Delete Classifier using Name == Failed")
            return 0
        if self.gbpverify.gbp_classif_verify(1, 'grppol_pc', id=cls_uuid) != 0:
            self._log.info(
                "\n## Step 3B: Verify Classifier is Deleted using "
                "-show option == Failed")
            return 0
        self._log.info("\n## TESTCASE_GBP_PC_FUNC_2: PASSED")

    def test_gbp_pc_func_3(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_PC_FUNC_3: UPDATE/VERIFY/DELETE/VERIFY ALL "
            "ATTRIB of a POLICY CLASSIFIER @ ONCE \n"
            "TEST_STEP::\n"
            "Create Policy Classifier using Default param values\n"
            "Update All the Polciy Classifier's editable params at one shot\n"
            "Verify the Policy Classifier's attributes & values, show "
            "& list cmds\n"
            "Delete the Policy Classifier\n"
            "Verify Policy Classifier successfully deleted\n"
            "##############################################################\n")
        # Testcase work-flow starts
        self._log.info(
            "\n## Step 1: Create Classifier with default attrib vals ##\n")
        cls_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'classifier', self.cls_name)
        if cls_uuid != 0:
            self._log.info(
                "\n## Step 1: Create Classifier Passed, UUID == %s\n" %
                (cls_uuid))
        else:
            self._log.info("\n## Step 1: Create Classifier == Failed")
            return 0
        self._log.info(
            "\n###########################################\n"
            "## Step 2: Update Policy Classifier Attributes ##\n"
            "## protocol, port-range,name,direction,description ##\n"
            "#################################################\n")
        if self.gbpcfg.gbp_policy_cfg_all(
                2,
                'classifier',
                cls_uuid,
                name='grppol_pc',
                protocol='tcp',
                direction='bi',
                port_range='22:1022',
                description="'For devstack demo'"):
            self._log.info(
                "\n## Step 2: Updating Policy Classifier's Attributes "
                "name,protocol,port-range,name,direction,description, Passed")
        else:
            self._log.info(
                "\n## Step 2: Updating Policy Classifier's Attributes "
                "name,protocol,port-range,name,direction,description, Failed")
            return 0

        if self.gbpverify.gbp_classif_verify(
                0, 'grppol_pc', cls_uuid, 'tcp', '22:1022', 'bi') == 0:
            self._log.info(
                "\n## Step 2A: Verify Policy Classifier Updated "
                "Attributes using -list option == Failed")
            return 0
        if self.gbpverify.gbp_classif_verify(
                1,
                'grppol_pc',
                id=cls_uuid,
                protocol='tcp',
                direction='bi',
                port_range='22:1022',
                description='For devstack demo') == 0:
            self._log.info(
                "\n## Step 2B: Verify Policy Classifier Updated "
                "Attributes using -show option == Failed")
            return 0
        if self.gbpcfg.gbp_policy_cfg_all(0, 'classifier', 'grppol_pc') == 0:
            self._log.info(
                "\n## Step 3: Delete Classifier using Name == Failed")
            return 0
        if self.gbpverify.gbp_classif_verify(1, 'grppol_pc', id=cls_uuid) != 0:
            self._log.info(
                "\n## Step 3B: Verify Classifier is Deleted using "
                "-show option == Failed")
            return 0
        self._log.info("\n## TESTCASE_GBP_PC_FUNC_3: PASSED")

    def test_gbp_pc_func_4(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_PC_FUNC_4: CREATE/SHARE/DELETE/ POLICY CLASSIFIER "
            "among MULTIPLE POLICY RULES \n"
            "TEST_STEP::\n"
            "Create and Verify Policy Classifier with valued attributes\n"
            "Create multiple(n=10) Policy Rules referencing the same "
            "Policy Classifier\n"
            "Verify the Policy Classifier is referenced in all "
            "configured Policy Rules\n"
            "Delete the Policy Classifier, verify it's deletion "
            "fails until all Policy Rules are deleted\n"
            "Verify Policy Classifier successfully deleted\n"
            "##############################################################\n")
        # Testcase work-flow starts
        self._log.info(
            "\n## Step 0: Creating a Policy Action needed for this test\n")
        act_uuid = self.gbpcfg.gbp_policy_cfg_all(1, 'action', self.act_name)
        if act_uuid == 0:
            self._log.info("\n## Testcase_gbp_pc_func_4: ABORTED\n")
            os._exit(1)
        self._log.info(
            "\n## Step 1: Create and Verify Classifier with valued "
            "attrib ##\n")
        cls_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1,
            'classifier',
            self.cls_name,
            protocol='tcp',
            direction='bi',
            port_range='22:1022',
            description="'For devstack demo'")
        if cls_uuid == 0:
            self._log.info("\n## Step 1A: Create Classifier == Failed")
            return 0
        if self.gbpverify.gbp_classif_verify(
                1,
                self.cls_name,
                id=cls_uuid,
                protocol='tcp',
                direction='bi',
                port_range='22:1022',
                description='For devstack demo') == 0:
            self._log.info(
                "\n## Step 1B: Verify Policy Classifier Attributes using "
                "-show option == Failed")
            return 0
        self._log.info(
            "\n## Step 2: Create Multiple Policy Rules, each referencing "
            "the same classifier ##\n")
        for n in range(1, 11):
            if self.gbpcfg.gbp_policy_cfg_all(
                    1, 'rule', 'grppol_pr_%s' %
                    (n), classifier=cls_uuid, action=act_uuid) == 0:
                self._log.info(
                    "\n## Step 2A: Policy Rule grppol_pr_%s creation, Failed" %
                    (n))
                return 0
            if self.gbpverify.gbp_policy_verify_all(
                    1,
                    'rule',
                    'grppol_pr_%s' %
                    (n),
                    policy_classifier_id=cls_uuid,
                    policy_actions=act_uuid) == 0:
                self._log.info(
                    "\n## Step 2B: Policy Rule grppol_pr_%s referencing "
                    "same classifier, Failed ##\n" %
                    (n))
                return 0
        self._log.info(
            "\n## Step 3: Delete Policy Classifier and Policy Rule and "
            "verify deletion fails ##")
        for i in range(1, 11):
            if self.gbpcfg.gbp_policy_cfg_all(0, 'classifier', cls_uuid) != 0:
                self._log.info(
                    "\n## Step 3A: Referenced Policy Classifier's "
                    "deletion DID NOT fail ##")
                return 0
            if self.gbpcfg.gbp_policy_cfg_all(
                    0, 'rule', 'grppol_pr_%s' %
                    (i)) == 0:
                self._log.info(
                    "\n## Step 3B: Referencing Policy Rule's deletion, "
                    "Failed ##")
                return 0
        self._log.info(
            "\n## Step 4: Deletion of Policy Classifier, all referencing "
            "Policy Rules have been deleted ##")
        if self.gbpcfg.gbp_policy_cfg_all(0, 'classifier', cls_uuid) == 0:
            self._log.info(
                "\n## Step 4A: Policy Classifier's deletion, Failed ##")
            return 0
        if self.gbpverify.gbp_classif_verify(1, 'grppol_pc', id=cls_uuid) != 0:
            self._log.info(
                "\n## Step 4B: Verify Classifier is Deleted, Failed")
            return 0
        self._log.info("\n## TESTCASE_GBP_PC_FUNC_4: PASSED")

if __name__ == '__main__':
    main()



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""gbp_classifiers_actions_rules

Revision ID: 4ae51f13395a
Create Date: 2014-07-30 14:16:05.660561

"""

# revision identifiers, used by Alembic.
revision = '4ae51f13395a'
down_revision = '1bf7555fa01a'


from alembic import op
from neutron_lib import constants
import sqlalchemy as sa

from gbpservice.neutron.services.grouppolicy.common import (
    constants as gp_constants)


def upgrade(active_plugins=None, options=None):
    op.create_table(
        'gp_policy_classifiers',
        sa.Column('id', sa.String(36), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('name', sa.String(length=50), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('protocol', sa.Enum(constants.PROTO_NAME_TCP,
                                      constants.PROTO_NAME_UDP,
                                      constants.PROTO_NAME_ICMP,
                                      name="protocol_type"),
                  nullable=True),
        sa.Column('port_range_min', sa.Integer),
        sa.Column('port_range_max', sa.Integer),
        sa.Column('direction', sa.Enum(gp_constants.GP_DIRECTION_IN,
                                       gp_constants.GP_DIRECTION_OUT,
                                       gp_constants.GP_DIRECTION_BI,
                                       name='direction'),
                  nullable=True),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'gp_policy_rules',
        sa.Column('id', sa.String(36), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('name', sa.String(length=50), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('enabled', sa.Boolean),
        sa.Column('policy_classifier_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['policy_classifier_id'],
                                ['gp_policy_classifiers.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'gp_policy_actions',
        sa.Column('id', sa.String(36), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('name', sa.String(length=50), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),

        sa.Column('action_type', sa.Enum(gp_constants.GP_ACTION_ALLOW,
                                         gp_constants.GP_ACTION_REDIRECT,
                                         name='action_type'),
                  nullable=True),
        sa.Column('action_value', sa.String(36), nullable=True),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'gp_policy_rule_action_associations',
        sa.Column('policy_rule_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['policy_rule_id'],
                                ['gp_policy_rules.id'], ondelete='CASCADE'),
        sa.Column('policy_action_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['policy_action_id'],
                                ['gp_policy_actions.id'], ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('policy_rule_id', 'policy_action_id'))


def downgrade(active_plugins=None, options=None):
    op.drop_table('gp_policy_rule_action_associations')
    op.drop_table('gp_policy_rules')
    op.drop_table('gp_policy_classifiers')
    op.drop_table('gp_policy_actions')



# Copyright (c) 2017 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from aim.api import resource as aim_resource
from neutron.db.models import address_scope as as_db
from neutron.db import models_v2
from neutron_lib.db import model_base
import sqlalchemy as sa
from sqlalchemy import orm


class AddressScopeMapping(model_base.BASEV2):
    __tablename__ = 'apic_aim_address_scope_mappings'

    scope_id = sa.Column(
        sa.String(36), sa.ForeignKey('address_scopes.id', ondelete='CASCADE'),
        primary_key=True)

    address_scope = orm.relationship(
        as_db.AddressScope, lazy='joined',
        backref=orm.backref(
            'aim_mapping', lazy='joined', uselist=False, cascade='delete'))

    vrf_name = sa.Column(sa.String(64))
    vrf_tenant_name = sa.Column(sa.String(64))
    vrf_owned = sa.Column(sa.Boolean, nullable=False)


class NetworkMapping(model_base.BASEV2):
    __tablename__ = 'apic_aim_network_mappings'

    network_id = sa.Column(
        sa.String(36), sa.ForeignKey('networks.id', ondelete='CASCADE'),
        primary_key=True)

    network = orm.relationship(
        models_v2.Network, lazy='joined',
        backref=orm.backref(
            'aim_mapping', lazy='joined', uselist=False, cascade='delete'))

    bd_name = sa.Column(sa.String(64))
    bd_tenant_name = sa.Column(sa.String(64))

    epg_name = sa.Column(sa.String(64))
    epg_app_profile_name = sa.Column(sa.String(64))
    epg_tenant_name = sa.Column(sa.String(64))

    l3out_name = sa.Column(sa.String(64))
    l3out_ext_net_name = sa.Column(sa.String(64))
    l3out_tenant_name = sa.Column(sa.String(64))

    vrf_name = sa.Column(sa.String(64))
    vrf_tenant_name = sa.Column(sa.String(64))


class DbMixin(object):
    def _add_address_scope_mapping(self, session, scope_id, vrf,
                                   vrf_owned=True, update_scope=True):
        mapping = AddressScopeMapping(
            scope_id=scope_id,
            vrf_name=vrf.name,
            vrf_tenant_name=vrf.tenant_name,
            vrf_owned=vrf_owned)
        session.add(mapping)
        if update_scope:
            # The AddressScope instance should already be in the
            # session cache, so this should not add another DB
            # roundtrip. It needs to be updated in case something
            # within the same transaction tries to access its
            # aim_mapping relationship after retrieving the
            # AddressScope record from the session cache.
            scope = (session.query(as_db.AddressScope).
                     filter_by(id=scope_id).
                     one_or_none())
            scope.aim_mapping = mapping
        return mapping

    def _get_address_scope_mapping(self, session, scope_id):
        return (session.query(AddressScopeMapping).
                filter_by(scope_id=scope_id).
                one_or_none())

    def _get_address_scope_mappings_for_vrf(self, session, vrf):
        return (session.query(AddressScopeMapping).
                filter_by(vrf_tenant_name=vrf.tenant_name,
                          vrf_name=vrf.name).
                all())

    def _get_address_scopes_owning_vrf(self, session, vrf):
        return (session.query(as_db.AddressScope).
                join(AddressScopeMapping,
                     AddressScopeMapping.scope_id == as_db.AddressScope.id).
                filter(AddressScopeMapping.vrf_tenant_name ==
                       vrf.tenant_name,
                       AddressScopeMapping.vrf_name == vrf.name,
                       AddressScopeMapping.vrf_owned).
                order_by(as_db.AddressScope.ip_version).
                all())

    def _get_address_scope_vrf(self, mapping):
        return aim_resource.VRF(
            tenant_name=mapping.vrf_tenant_name,
            name=mapping.vrf_name)

    def _add_network_mapping(self, session, network_id, bd, epg, vrf,
                             ext_net=None, update_network=True):
        if not ext_net:
            mapping = NetworkMapping(
                network_id=network_id,
                bd_name=bd.name,
                bd_tenant_name=bd.tenant_name,
                epg_name=epg.name,
                epg_app_profile_name=epg.app_profile_name,
                epg_tenant_name=epg.tenant_name,
                vrf_name=vrf.name,
                vrf_tenant_name=vrf.tenant_name)
        else:
            mapping = NetworkMapping(
                network_id=network_id,
                l3out_name=ext_net.l3out_name,
                l3out_ext_net_name=ext_net.name,
                l3out_tenant_name=ext_net.tenant_name,
                vrf_name=vrf.name,
                vrf_tenant_name=vrf.tenant_name)
        session.add(mapping)
        if update_network:
            # The Network instance should already be in the session
            # cache, so this should not add another DB roundtrip. It
            # needs to be updated in case something within the same
            # transaction tries to access its aim_mapping relationship
            # after retrieving the Network record from the session
            # cache.
            net = (session.query(models_v2.Network).
                   filter_by(id=network_id).
                   one_or_none())
            net.aim_mapping = mapping
        return mapping

    def _get_network_mapping(self, session, network_id):
        return (session.query(NetworkMapping).
                filter_by(network_id=network_id).
                one_or_none())

    def _get_network_mapping_bulk(self, session, network_ids):
        return session.query(NetworkMapping).filter(
            NetworkMapping.network_id.in_(network_ids)).all()

    def _get_network_mappings_for_vrf(self, session, vrf):
        return (session.query(NetworkMapping).
                filter_by(vrf_tenant_name=vrf.tenant_name,
                          vrf_name=vrf.name).
                all())

    def _get_network_mappings_for_bd(self, session, bd):
        return (session.query(NetworkMapping).
                filter_by(bd_tenant_name=bd.tenant_name,
                          bd_name=bd.name).
                all())

    def _is_vrf_used_by_networks(self, session, vrf):
        return (session.query(NetworkMapping.network_id).
                filter_by(vrf_tenant_name=vrf.tenant_name,
                          vrf_name=vrf.name).
                first() is not None)

    def _get_network_bd(self, mapping):
        return aim_resource.BridgeDomain(
            tenant_name=mapping.bd_tenant_name,
            name=mapping.bd_name)

    def _get_network_epg(self, mapping):
        return aim_resource.EndpointGroup(
            tenant_name=mapping.epg_tenant_name,
            app_profile_name=mapping.epg_app_profile_name,
            name=mapping.epg_name)

    def _get_network_l3out(self, mapping):
        if not mapping:
            # REVISIT: Is this still needed now that
            # _add_network_mapping updates the Network instance's
            # aim_mapping? If so, the test should probably be moved to
            # the caller to make all these
            # _get_<neutron-resource>_<aim-resource> methods more
            # consistent.
            return None
        return aim_resource.L3Outside(
            tenant_name=mapping.l3out_tenant_name,
            name=mapping.l3out_name)

    def _get_network_l3out_ext_net(self, mapping):
        return aim_resource.ExternalNetwork(
            tenant_name=mapping.l3out_tenant_name,
            l3out_name=mapping.l3out_name, name=mapping.l3out_ext_net_name)

    def _get_network_l3out_default_ext_subnetv4(self, mapping):
        return aim_resource.ExternalSubnet(
            tenant_name=mapping.l3out_tenant_name,
            l3out_name=mapping.l3out_name,
            external_network_name=mapping.l3out_ext_net_name,
            cidr="0.0.0.0/0")

    def _get_network_l3out_default_ext_subnetv6(self, mapping):
        return aim_resource.ExternalSubnet(
            tenant_name=mapping.l3out_tenant_name,
            l3out_name=mapping.l3out_name,
            external_network_name=mapping.l3out_ext_net_name,
            cidr="::/0")

    def _get_network_vrf(self, mapping):
        return aim_resource.VRF(
            tenant_name=mapping.vrf_tenant_name,
            name=mapping.vrf_name)

    def _set_network_bd(self, mapping, bd):
        mapping.bd_tenant_name = bd.tenant_name
        mapping.bd_name = bd.name

    def _set_network_epg(self, mapping, epg):
        mapping.epg_tenant_name = epg.tenant_name
        mapping.epg_app_profile_name = epg.app_profile_name
        mapping.epg_name = epg.name

    def _set_network_l3out(self, mapping, l3out):
        mapping.l3out_tenant_name = l3out.tenant_name
        mapping.l3out_name = l3out.name

    def _set_network_vrf(self, mapping, vrf):
        mapping.vrf_tenant_name = vrf.tenant_name
        mapping.vrf_name = vrf.name



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import signal
import time

from oslo_service import service as oslo_service

from gbpservice.nfp.core import log as nfp_logging

LOG = nfp_logging.getLogger(__name__)
ProcessLauncher = oslo_service.ProcessLauncher

"""Worker process launcher.

    Derives the oslo process launcher to
    launch childrens with python multiprocessing
    as oppose to os.fork(), coz, communication
    is needed from parent->child not just the
    parallel execution.
"""


class NfpLauncher(ProcessLauncher):

    def __init__(self, conf):
        # Add SIGALARM to ignore_signals, because core
        # uses SIGALRM for watchdog, while oslo uses the
        # same for exit.
        # Signal handler is singleton class, changing here will
        # have global effect.
        self.signal_handler = oslo_service.SignalHandler()
        self.signal_handler._ignore_signals += ('SIGALRM',)
        self.signal_handler._signals_by_name = dict(
            (name, getattr(signal, name))
            for name in dir(signal)
            if name.startswith("SIG") and
            name not in self.signal_handler._ignore_signals)

        super(NfpLauncher, self).__init__(conf)

    def child(self, service, ppipe, cpipe, controller):
        service.parent_pipe = ppipe
        service.pipe = cpipe
        service.controller = controller
        self.launcher = self._child_process(service)
        while True:
            self._child_process_handle_signal()
            status, signo = self._child_wait_for_exit_or_signal(
                self.launcher)
            if not oslo_service._is_sighup_and_daemon(signo):
                self.launcher.wait()
                break
            self.launcher.restart()

        os._exit(status)

    def _start_child(self, wrap):
        if len(wrap.forktimes) > wrap.workers:
            # Limit ourselves to one process a second (over the period of
            # number of workers * 1 second). This will allow workers to
            # start up quickly but ensure we don't fork off children that
            # die instantly too quickly.
            if time.time() - wrap.forktimes[0] < wrap.workers:
                time.sleep(1)

            wrap.forktimes.pop(0)

        wrap.forktimes.append(time.time())

        pid = self.fork_child(wrap)

        message = "Started Child Process %d" % (pid)
        LOG.debug(message)

        wrap.children.add(pid)
        self.children[pid] = wrap

        return pid

    def fork_child(self, wrap):
        # Default use os.fork to create a child
        pid = os.fork()
        if pid == 0:
            self.launcher = self._child_process(wrap.service)
            while True:
                self._child_process_handle_signal()
                status, signo = self._child_wait_for_exit_or_signal(
                    self.launcher)
                if not oslo_service._is_sighup_and_daemon(signo):
                    self.launcher.wait()
                    break
                self.launcher.restart()

            os._exit(status)
        return pid



GBP_PLURALS = {}


def register_plurals(plural_mappings):
    for plural, single in plural_mappings.items():
        GBP_PLURALS[single] = plural


def get_plural(single):
    return GBP_PLURALS.get(single)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions


class UnsupportedConfiguration(exceptions.BadRequest):
    message = _("Configuration %(conf)s for object of type %(type)s is not "
                "supported.")


class PortPairsDifferentNetworkInGroup(exceptions.BadRequest):
    message = _("For a Port Pair Group, ingress and egress networks should "
                "be the same across Port Pairs. Example of valid "
                "port pairs networks: [(N1, N2), (N1, N2), (N1, N2)]. "
                "Invalid Example: [(N1, N2), (N1, N3), (N3, N2)]. "
                "Port Pair Group ID: %(id)s")


class PortPairsUnsupportedDomain(exceptions.BadRequest):
    message = _("Port Pair's ingress and egress port domains are unsupported "
                "Please check the AIMCTL HostDomainMappingV2 "
                "port pair ID: %(id)s supported domains: %(doms)s")


class PortPairsDifferentDomain(exceptions.BadRequest):
    message = _("Port Pair's ingress and egress port can't be in different "
                "domains. Please check the AIMCTL HostDomainMappingV2 "
                "port pair ID: %(id)s")


class PortPairsNoUniqueDomain(exceptions.BadRequest):
    message = _("Port Pair's ingress and egress port domains can't be "
                "unequivocally identified. Please check the AIMCTL "
                "HostDomainMappingV2 port pair ID: %(id)s")


class PortPairsSameNetwork(exceptions.BadRequest):
    message = _("Port Pair's ingress and egress port can't be in the same "
                "network. port pair ID: %(id)s")


class PortPairsInPortPairGroupDifferentDomain(exceptions.BadRequest):
    message = _("Port Pairs in the same PPG must be in the same domain, "
                "Please check the AIMCTL HostDomainMappingV2 "
                "port pair ID: %(id)s")


class BadFlowClassifier(exceptions.BadRequest):
    message = _("The following L7 parameters must be configured on Flow "
                "Classifiers when using AIM driver: %(params)s")


class FlowClassifierSameSrcDstNetworks(exceptions.BadRequest):
    message = _("Source and Destination networks must be different in Flow "
                "Classifier if not SVI.")


class RouterIDInUseByFlowClassifier(exceptions.BadRequest):
    message = _("Router %(router_id)s is in use by Flow Classifier "
                "%(flowc_id)s.")


class NoL3OutAssociatedToFlowcExternalNetwork(exceptions.BadRequest):
    message = _("Cannot map flow classifier %(id)s, either its source or "
                "destination network is external but has no L3Outside "
                "associated to it.")


class NoPhysicalDomainSpecifiedInServiceEPG(exceptions.BadRequest):
    message = _("No Physical Domain is specified in service EPG %(epg_id)s. ")


class MultipleVRFsDetectedInPortChain(exceptions.BadRequest):
    message = _("Port Chain %(id)s spans across multiple VRFs. All providers, "
                "consumers, and service BDs have to be in the same VRF.")


class FlowClassifierSrcDstNetworksDifferentTenants(exceptions.BadRequest):
    message = _("Source and Destination networks for flow classifier %(id)s "
                "are in different tenants. This is currently unsupported.")


class NetworkInUseByFlowClassifiers(exceptions.BadRequest):
    message = _("Cannot delete network in use by classifiers %(ids)s")


class ServiceNetworkBadType(exceptions.BadRequest):
    message = _("Service networks can't be SVI or External. "
                "Port Pair ID: %(id)s")


class ConflictingNetworksDetectedInPortChain(exceptions.BadRequest):
    message = _("Port Pair Groups in Port Chain cannot share any network. "
                "%(id)s")


class DefaultExternalNetworkNotFound(exceptions.NotFound):
    message = _("Default External Network not found for SVI network "
                "%(id)s.")


class TooManyPPGsPerChainError(exceptions.BadRequest):
    message = _("The max number of PPGs per chain supported is %(maxn)s.")


class FlowClassifierSameSrcDstSVISameSubnet(exceptions.BadRequest):
    message = _("Source and Destination networks must have different subnets "
                "in Flow Classifier if SVI.")


class FlowClassifierSameSubnet(exceptions.BadRequest):
    message = _("Source and Destination cidrs must be different.")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import subprocess
import sys


def run_func_neg():
    """
    Method to manipulate needed files
    """
    # Assumption is all files are in current directory
    # if 'Ubuntu' in platform.linux_distribution():
    #    directory = "/usr/local/lib/python2.7/dist-packages/gbpfunctests/"
    # else:
    #   directory = "/usr/lib/python2.7/site-packages/gbpfunctests/"  # in RHEL
    cmd_list = ["sudo sh -c 'cat /dev/null > test_results.txt'",
                "sudo sh -c 'cat /dev/null > func_neg.txt'",
                "sudo sh -c 'ls *_func*.py | grep -v shared > func_neg.txt'",
                "sudo sh -c 'ls *_neg.py >> func_neg.txt'",
                "sudo chmod 777 *"]
    for cmd in cmd_list:
        commands.getoutput(cmd)
    return "func_neg.txt"


def main():
    """
    Test Execution
    Generating Results
    """
    usage = "Usage: python suite_run.py <'aci' or 'upstream'>"
    try:
        flag = sys.argv[1]
    except Exception:
        print ('%s' % (usage))
        sys.exit(1)
    fname = run_func_neg()
    num_lines = sum(1 for line in open(fname))
    print ("\nNumber of Functional Test Scripts to execute = %s" % (num_lines))
    with open(fname) as f:
        for i, l in enumerate(f, 1):
            print ("Functional Test Script to execute now == %s" % l)
            # Assumption: test-scripts are executable from any location
            # Reading the line from text file, also reads trailing \n, hence we
            # need to strip
            cmd = 'python %s %s' % (l.strip(), flag)
            print (cmd)
            # out=getoutput(cmd)
            subprocess.call(cmd, shell=True)
    f = open("test_results.txt")
    contents = f.read()
    f.close()
    print (contents)
    print ("\n\nTotal Number of TestCases Executed= %s" % (
        contents.count("TESTCASE_GBP_")))
    print ("\n\nNumber of TestCases Passed= %s" % (contents.count("PASSED")))
    print ("\n\nNumber of TestCases Failed= %s" % (contents.count("FAILED")))
    if contents.count("FAILED") > 0:
        sys.exit(1)
    else:
        return 0

if __name__ == "__main__":
    main()



# Copyright 2011 VMware, Inc., 2014 A10 Networks
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Routines for configuring Octavia Haproxy driver
"""

from oslo_config import cfg

from gbpservice.nfp.core import log as nfp_logging

LOG = nfp_logging.getLogger(__name__)

haproxy_amphora_opts = [
    cfg.StrOpt('base_path',
               default='/var/lib/octavia',
               help=_('Base directory for amphora files.')),
    cfg.StrOpt('base_cert_dir',
               default='/var/lib/octavia/certs',
               help=_('Base directory for cert storage.')),
    cfg.StrOpt('haproxy_template', help=_('Custom haproxy template.')),
    cfg.IntOpt('connection_max_retries',
               default=10,
               help=_('Retry threshold for connecting to amphorae.')),
    cfg.IntOpt('connection_retry_interval',
               default=5,
               help=_('Retry timeout between connection attempts in '
                      'seconds.')),
    cfg.StrOpt('haproxy_stick_size', default='10k',
               help=_('Size of the HAProxy stick table. Accepts k, m, g '
                      'suffixes.  Example: 10k')),

    # REST server
    cfg.IPOpt('bind_host', default='0.0.0.0',
              help=_("The host IP to bind to")),
    cfg.PortOpt('bind_port', default=9443,
                help=_("The port to bind to")),
    cfg.StrOpt('haproxy_cmd', default='/usr/sbin/haproxy',
               help=_("The full path to haproxy")),
    cfg.IntOpt('respawn_count', default=2,
               help=_("The respawn count for haproxy's upstart script")),
    cfg.IntOpt('respawn_interval', default=2,
               help=_("The respawn interval for haproxy's upstart script")),
    cfg.FloatOpt('rest_request_conn_timeout', default=10,
                 help=_("The time in seconds to wait for a REST API "
                        "to connect.")),
    cfg.FloatOpt('rest_request_read_timeout', default=60,
                 help=_("The time in seconds to wait for a REST API "
                        "response.")),
    # REST client
    cfg.StrOpt('client_cert', default='/etc/octavia/certs/client.pem',
               help=_("The client certificate to talk to the agent")),
    cfg.StrOpt('server_ca', default='/etc/octavia/certs/server_ca.pem',
               help=_("The ca which signed the server certificates")),
    cfg.BoolOpt('use_upstart', default=True,
                help=_("If False, use sysvinit.")),
]

certificate_opts = [
    cfg.StrOpt('cert_manager',
               default='local_cert_manager',
               help='Name of the cert manager to use'),
    cfg.StrOpt('cert_generator',
               default='local_cert_generator',
               help='Name of the cert generator to use'),
]

# Options only used by the amphora agent
amphora_agent_opts = [
    cfg.StrOpt('agent_server_ca', default='/etc/octavia/certs/client_ca.pem',
               help=_("The ca which signed the client certificates")),
    cfg.StrOpt('agent_server_cert', default='/etc/octavia/certs/server.pem',
               help=_("The server certificate for the agent.py server "
                      "to use")),
    cfg.StrOpt('agent_server_network_dir',
               default='/etc/network/interfaces.d/',
               help=_("The directory where new network interfaces "
                      "are located")),
    cfg.StrOpt('agent_server_network_file',
               help=_("The file where the network interfaces are located. "
                      "Specifying this will override any value set for "
                      "agent_server_network_dir.")),
    # Do not specify in octavia.conf, loaded at runtime
    cfg.StrOpt('amphora_id', help=_("The amphora ID.")),
]

keepalived_vrrp_opts = [
    cfg.IntOpt('vrrp_advert_int',
               default=1,
               help=_('Amphora role and priority advertisement interval '
                      'in seconds.')),
    cfg.IntOpt('vrrp_check_interval',
               default=5,
               help=_('VRRP health check script run interval in seconds.')),
    cfg.IntOpt('vrrp_fail_count',
               default=2,
               help=_('Number of successive failure before transition to a '
                      'fail state.')),
    cfg.IntOpt('vrrp_success_count',
               default=2,
               help=_('Number of successive failure before transition to a '
                      'success state.')),
    cfg.IntOpt('vrrp_garp_refresh_interval',
               default=5,
               help=_('Time in seconds between gratuitous ARP announcements '
                      'from the MASTER.')),
    cfg.IntOpt('vrrp_garp_refresh_count',
               default=2,
               help=_('Number of gratuitous ARP announcements to make on '
                      'each refresh interval.'))

]

# Register the configuration options
cfg.CONF.register_opts(amphora_agent_opts, group='amphora_agent')
cfg.CONF.register_opts(certificate_opts, group='certificates')
cfg.CONF.register_opts(haproxy_amphora_opts, group='haproxy_amphora')
cfg.CONF.register_opts(keepalived_vrrp_opts, group='keepalived_vrrp')
CONF = cfg.CONF



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice._i18n import _LE
from gbpservice.contrib.nfp.config_orchestrator.common import (
    topics as a_topics)
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.lib import transport
from gbpservice.nfp.orchestrator.openstack import openstack_driver

from neutron.common import constants as n_constants
from neutron.common import rpc as n_rpc
from neutron.common import topics as n_topics

import oslo_messaging as messaging

LOG = nfp_logging.getLogger(__name__)


def prepare_request_data(context, resource, resource_type,
                         resource_data, service_vendor=None):

    request_data = {'info': {
        'context': context,
        'service_type': resource_type,
        'service_vendor': service_vendor
    },

        'config': [{
            'resource': resource,
            'resource_data': resource_data
        }]
    }

    return request_data


def _filter_data(routers, networks, filters):
    # filter routers and networks data and formulate
    # dictionary of subnets, routers and ports for the
    # given tenant.
    tenant_id = filters['tenant_id'][0]
    _filtered_routers = []
    _filtered_subnets = []
    _filtered_ports = []
    _filtered_networks = []
    for router in routers:
        if router['tenant_id'] == tenant_id:
            _filtered_routers.append({'id': router['id']})
    for network in networks:
        if network['tenant_id'] == tenant_id:
            subnets = network['subnets']
            ports = network['ports']
            _filtered_networks.append(
                {'id': network['id'],
                 'tenant_id': network['tenant_id'],
                 'provider:segmentation_id': network[
                     'provider:segmentation_id'],
                 'provider:network_type': network[
                     'provider:network_type'],
                 'shared': network['shared'],
                 'router:external': network['router:external']})
            for subnet in subnets:
                if subnet['tenant_id'] == tenant_id:
                    _filtered_subnets.append(
                        {'id': subnet['id'],
                         'cidr': subnet['cidr'],
                         'gateway_ip': subnet['gateway_ip'],
                         'network_id': subnet['network_id']})
            for port in ports:
                if port['tenant_id'] == tenant_id:
                    _filtered_ports.append(
                        {'id': port['id'],
                         'fixed_ips': port['fixed_ips'],
                         'binding:host_id': port['binding:host_id'],
                         'network_id': port['network_id']})

    return {'subnets': _filtered_subnets,
            'routers': _filtered_routers,
            'ports': _filtered_ports,
            'networks': _filtered_networks}


def get_core_context(context, filters, config):
    #routers = get_routers(context, config.host)
    routers = []
    networks = get_networks(context, config)
    return _filter_data(routers, networks, filters)


def get_routers(context, host):
    target = messaging.Target(topic=n_topics.L3PLUGIN, version='1.0')
    client = n_rpc.get_client(target)
    cctxt = client.prepare()
    return cctxt.call(context, 'sync_routers', host=host,
                      router_ids=None)


def get_dhcp_agent_host(config):
    try:
        neutronclient = openstack_driver.NeutronClient(config)
        keystoneclient = openstack_driver.KeystoneClient(config)
        token = keystoneclient.get_admin_token()
        filters = {'agent_type': 'DHCP agent', 'alive': True}
        agents = neutronclient.get_agents(token, filters)
        if agents:
            return agents[0].get('host', None)
    except Exception as exc:
        LOG.error(_LE("Failed to get dhcp agent host : %(exc)s"),
                  {'exc': exc})


def get_networks(context, config):
    host = get_dhcp_agent_host(config)
    if not host:
        return []
    target = messaging.Target(
        topic=n_topics.PLUGIN,
        namespace=n_constants.RPC_NAMESPACE_DHCP_PLUGIN,
        version='1.0')
    client = n_rpc.get_client(target)
    cctxt = client.prepare(version='1.1')
    return cctxt.call(context, 'get_active_networks_info',
                      host=host)


def _prepare_structure(network_function_details, ports_info,
                       mngmt_port_info, monitor_port_info):
    return {'nfi_ports_map': {
        network_function_details[
            'network_function_instance'][
            'id']: ports_info},
            'nfi_nfd_map': {
                network_function_details[
                    'network_function_instance'][
                    'id']: {
                    'nfd': network_function_details[
                        'network_function_device'],
                    'nfd_mgmt_port': mngmt_port_info,
                    'nfd_monitoring_port': None,
                    'nfd_monitoring_port_network': network_function_details[
                        'network_function_device'][
                            'monitoring_port_network']}},
            'nfi': [network_function_details['network_function_instance']],
            'nf': network_function_details['network_function']
            }


def get_network_function_details(context, network_function_id):
    network_function_details = None
    try:
        rpc_nso_client = transport.RPCClient(a_topics.NFP_NSO_TOPIC)
        network_function_details = rpc_nso_client.cctxt.call(
            context,
            'get_network_function_details',
            network_function_id=network_function_id)
        msg = (" %s " % (network_function_details))
        LOG.debug(msg)
        return network_function_details['network_function']

    except Exception as e:
        LOG.error(_LE("Failed to get network function details of "
                      "network_function_id %(network_function_id)s : %(ex)s "),
                  {'ex': e, 'network_function_id': network_function_id})


def get_network_function_map(context, network_function_id):
    request_data = None
    try:
        rpc_nso_client = transport.RPCClient(a_topics.NFP_NSO_TOPIC)
        nf_context = rpc_nso_client.cctxt.call(
            context,
            'get_network_function_context',
            network_function_id=network_function_id)

        network_function_details = nf_context['network_function_details']
        ports_info = nf_context['ports_info']
        mngmt_port_info = nf_context['mngmt_port_info']
        monitor_port_info = nf_context['monitor_port_info']

        request_data = _prepare_structure(network_function_details, ports_info,
                                          mngmt_port_info, monitor_port_info)
        msg = (" %s " % (request_data))
        LOG.debug(msg)
        return request_data
    except Exception as e:
        LOG.error(_LE("Failed to get network function map of "
                      "network_function_id %(network_function_id)s : %(ex)s "),
                  {'ex': e, 'network_function_id': network_function_id})
        return request_data



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import oslo_serialization.jsonutils as jsonutils

from neutron.common import rpc as n_rpc
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging
import pecan
import pika

from gbpservice.nfp.pecan import base_controller

LOG = logging.getLogger(__name__)
n_rpc.init(cfg.CONF)


class Controller(base_controller.BaseController):
    """Implements all the APIs Invoked by HTTP requests.

    Implements following HTTP methods.
        -get
        -post
        -put
    According to the HTTP request received from config-agent this class make
    call/cast to configurator and return response to config-agent

    """

    def __init__(self, method_name):
        try:
            self.method_name = method_name
            self.services = pecan.conf['cloud_services']
            self.rpc_routing_table = {}
            for service in self.services:
                self._entry_to_rpc_routing_table(service)

            configurator_notifications = self.services[0]['notifications']
            self.rmqconsumer = RMQConsumer(configurator_notifications['host'],
                                           configurator_notifications['queue']
                                           )
            super(Controller, self).__init__()
        except Exception as err:
            msg = (
                "Failed to initialize Controller class  %s." %
                str(err).capitalize())
            LOG.error(msg)

    def _entry_to_rpc_routing_table(self, service):
        """Prepares routing table based on the uservice configuration.
           This routing table is used to route the rpcs to all interested
           uservices. Key used for routing is the uservice[apis].

        :param uservice
        e.g uservice = {'service_name': 'configurator',
                        'topic': 'configurator',
                        'reporting_interval': '10',  # in seconds
                        'apis': ['CONFIGURATION', 'EVENT']
                        }
        Returns: None

        Prepares: self.rpc_routing_table
            e.g self.rpc_routing_table = {'CONFIGURATION': [rpc_client, ...],
                                          'EVENT': [rpc_client, ...]
                                          }
        """
        for api in service['apis']:
            if api not in self.rpc_routing_table:
                self.rpc_routing_table[api] = []

            self.rpc_routing_table[api].append(CloudService(**service))

    @pecan.expose(method='GET', content_type='application/json')
    def get(self):
        """Method of REST server to handle request get_notifications.

        This method send an RPC call to configurator and returns Notification
        data to config-agent

        Returns: Dictionary that contains Notification data

        """

        try:
            if self.method_name == 'get_notifications':
                notification_data = self.rmqconsumer.pull_notifications()
                msg = ("NOTIFICATION_DATA sent to config_agent %s"
                       % notification_data)
                LOG.info(msg)
                return jsonutils.dumps(notification_data)

        except Exception as err:
            pecan.response.status = 400
            msg = ("Failed to handle request=%s. Reason=%s."
                   % (self.method_name, str(err).capitalize()))
            LOG.error(msg)
            error_data = self._format_description(msg)
            return jsonutils.dumps(error_data)

    @pecan.expose(method='POST', content_type='application/json')
    def post(self, **body):
        """Method of REST server to handle all the post requests.

        This method sends an RPC cast to configurator according to the
        HTTP request.

        :param body: This method excepts dictionary as a parameter in HTTP
        request and send this dictionary to configurator with RPC cast.

        Returns: None

        """

        try:
            body = None
            if pecan.request.is_body_readable:
                body = pecan.request.json_body

            routing_key = body.pop("routing_key", "CONFIGURATION")
            for uservice in self.rpc_routing_table[routing_key]:
                uservice.rpcclient.cast(self.method_name, body)
                msg = ('Sent RPC to %s' % (uservice.topic))
                LOG.info(msg)

            msg = ("Successfully served HTTP request %s" % self.method_name)
            LOG.info(msg)

        except Exception as err:
            pecan.response.status = 400
            msg = ("Failed to serve HTTP post request %s %s."
                   % (self.method_name, str(err).capitalize()))
            # extra_import = ("need to remove this import %s" % config)
            # LOG.debug(extra_import)
            LOG.error(msg)
            error_data = self._format_description(msg)
            return jsonutils.dumps(error_data)

    @pecan.expose(method='PUT', content_type='application/json')
    def put(self, **body):
        """Method of REST server to handle all the put requests.

        This method sends an RPC cast to configurator according to the
        HTTP request.

        :param body: This method excepts dictionary as a parameter in HTTP
        request and send this dictionary to configurator with RPC cast.

        Returns: None

        """
        try:
            body = None
            if pecan.request.is_body_readable:
                body = pecan.request.json_body

            routing_key = body.pop("routing_key", "CONFIGURATION")
            for uservice in self.rpc_routing_table[routing_key]:
                uservice.rpcclient.cast(self.method_name, body)
                msg = ('Sent RPC to %s' % (uservice.topic))
                LOG.info(msg)
            msg = ("Successfully served HTTP request %s" % self.method_name)
            LOG.info(msg)

        except Exception as err:
            pecan.response.status = 400
            msg = ("Failed to serve HTTP put request %s %s."
                   % (self.method_name, str(err).capitalize()))
            LOG.error(msg)
            error_data = self._format_description(msg)
            return jsonutils.dumps(error_data)

    def _format_description(self, msg):
        """This methgod formats error description.

        :param msg: An error message that is to be formatted

        Returns: error_data dictionary
        """

        error_data = {'failure_desc': {'msg': msg}}
        return error_data


class RPCClient(object):
    """Implements call/cast methods used in REST Controller.

    Implements following methods.
        -call
        -cast
    This class send an RPC call/cast to configurator according to the data sent
    by Controller class of REST server.

     """

    API_VERSION = '1.0'

    def __init__(self, topic):

        self.topic = topic
        target = oslo_messaging.Target(
            topic=self.topic,
            version=self.API_VERSION)
        self.client = n_rpc.get_client(target)

    def call(self, method_name):
        """Method for sending call request on behalf of REST Controller.

        This method sends an RPC call to configurator.

        Returns: Notification data sent by configurator.

        """
        cctxt = self.client.prepare(version=self.API_VERSION,
                                    topic=self.topic)
        return cctxt.call(self, method_name)

    def cast(self, method_name, request_data):
        """Method for sending cast request on behalf of REST Controller.

        This method sends an RPC cast to configurator according to the
        method_name passed by COntroller class of REST server.

        :param method_name:method name can be any of the following.


        Returns: None.

        """
        cctxt = self.client.prepare(version=self.API_VERSION,
                                    topic=self.topic)

        return cctxt.cast(self,
                          method_name,
                          request_data=request_data)

    def to_dict(self):
        """This function return empty dictionary.

        For making RPC call/cast it internally requires context class that
        contains to_dict() function. Here we are sending context inside
        request data so we are passing class itself as a context that
        contains to_dict() function.

        Returns: Dictionary.

        """
        return {}


class CloudService(object):
    """ CloudService keeps all information of uservice along with initialized
        RPCClient object using which rpc is routed to over the cloud service.
    """

    def __init__(self, **kwargs):
        self.service_name = kwargs.get('service_name')
        self.topic = kwargs.get('topic')
        self.reporting_interval = kwargs.get('reporting_interval')
        self.rpcclient = RPCClient(topic=self.topic)


class RMQConsumer(object):
    """RMQConsumer for over the cloud services.

    This class access rabbitmq's 'configurator-notifications' queue
    to pull all the notifications came from over the cloud services.

    """

    def __init__(self, rabbitmq_host, queue):
        self.rabbitmq_host = rabbitmq_host
        self.queue = queue
        self.create_connection()

    def create_connection(self):
        try:
            self.connection = pika.BlockingConnection(
                                    pika.ConnectionParameters
                                    (host=self.rabbitmq_host))
        except Exception as e:
            msg = ("Failed to create rmq connection %s" % (e))
            LOG.error(msg)

    def _fetch_data_from_wrapper_strct(self, oslo_notifications):
        notifications = []
        for oslo_notification_data in oslo_notifications:
            notification_data = jsonutils.loads(
                oslo_notification_data["oslo.message"]
            )["args"]["notification_data"]
            notifications.extend(notification_data)
        return notifications

    def pull_notifications(self):
        notifications = []
        msgs_acknowledged = False
        try:
            self.channel = self.connection.channel()
            self.queue_declared = self.channel.queue_declare(queue=self.queue,
                                                             durable=True)
            self.channel.queue_bind(self.queue, 'openstack')
            pending_msg_count = self.queue_declared.method.message_count
            log = ('[notifications queue:%s, pending notifications:%s]'
                   % (self.queue, pending_msg_count))
            LOG.info(log)
            for i in range(pending_msg_count):
                method, properties, body = self.channel.basic_get(self.queue)
                notifications.append(jsonutils.loads(body))

            # Acknowledge all messages delivery
            if pending_msg_count > 0:
                self.channel.basic_ack(delivery_tag=method.delivery_tag,
                                       multiple=True)
                msgs_acknowledged = True

            self.channel.close()
            return self._fetch_data_from_wrapper_strct(notifications)

        except pika.exceptions.ConnectionClosed:
            msg = ("Caught ConnectionClosed exception.Creating new connection")
            LOG.error(msg)
            self.create_connection()
            return self._fetch_data_from_wrapper_strct(notifications)
        except pika.exceptions.ChannelClosed:
            msg = ("Caught ChannelClosed exception.")
            LOG.error(msg)
            if msgs_acknowledged is False:
                return self.pull_notifications()
            else:
                return self._fetch_data_from_wrapper_strct(notifications)



# The following are imported at the beginning to ensure that the
# patches are applied before any of the modules save a reference to
# the functions being patched. The order is also important.
from gbpservice.neutron.extensions import patch  # noqa

from gbpservice.neutron.plugins.ml2plus import patch_neutron  # noqa



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron.callbacks import registry
from neutron.extensions import address_scope
from neutron.extensions import l3
from neutron.extensions import securitygroup as ext_sg
from neutron import manager
from neutron.notifiers import nova
from neutron.plugins.common import constants as pconst
from neutron import quota
from neutron_lib import exceptions as n_exc
from oslo_log import log as logging
from oslo_utils import excutils

from gbpservice._i18n import _LE
from gbpservice._i18n import _LW
from gbpservice.neutron.extensions import group_policy as gp_ext
from gbpservice.neutron.extensions import servicechain as sc_ext
from gbpservice.neutron.services.grouppolicy.common import exceptions as exc

LOG = logging.getLogger(__name__)


def get_outer_transaction(transaction):
    if not transaction:
        return
    if not transaction._parent:
        return transaction
    else:
        return get_outer_transaction(transaction._parent)


# Note: QUEUE_OUT_OF_PROCESS_NOTIFICATIONS can be set to
# True only by drivers which use the ml2plus neutron plugin
QUEUE_OUT_OF_PROCESS_NOTIFICATIONS = False
NOVA_NOTIFIER_METHOD = 'send_network_change'
DHCP_NOTIFIER_METHOD = 'notify'
NOTIFIER_REF = 'notifier_object_reference'
NOTIFIER_METHOD = 'notifier_method_name'
NOTIFICATION_ARGS = 'notification_args'
REGISTRY_RESOURCE = 'registry_resource'
REGISTRY_EVENT = 'registry_event'
REGISTRY_TRIGGER = 'registry_trigger'
# Add known agent RPC notifiers here. These notifiers will be invoked
# in a delayed manner after the outermost transaction that initiated
# the notification has completed.
# These module names/prefixes are mutually exclusive from the
# notifiers/notifications that are handled in process.
OUT_OF_PROCESS_NOTIFICATIONS = ['neutron.api.rpc.agentnotifiers',
                                'neutron.notifiers.nova', 'opflexagent.rpc']


def _enqueue(session, transaction_key, entry):
    if transaction_key not in session.notification_queue:
        session.notification_queue[transaction_key] = [entry]
    else:
        session.notification_queue[transaction_key].append(entry)


def _queue_notification(session, transaction_key, notifier_obj,
                        notifier_method, args):
    entry = {NOTIFIER_REF: notifier_obj, NOTIFIER_METHOD: notifier_method,
             NOTIFICATION_ARGS: args}
    _enqueue(session, transaction_key, entry)


def _queue_registry_notification(session, transaction_key, resource,
                                 event, trigger, **kwargs):
    entry = {REGISTRY_RESOURCE: resource, REGISTRY_EVENT: event,
             REGISTRY_TRIGGER: trigger, NOTIFICATION_ARGS: kwargs}
    _enqueue(session, transaction_key, entry)


def send_or_queue_notification(session, transaction_key, notifier_obj,
                               notifier_method, args):
    rname = ''
    if notifier_method == NOVA_NOTIFIER_METHOD:
        # parse argument like "create_subnetpool"
        rname = args[0].split('_', 1)[1]
        event_name = 'after_' + args[0].split('_')[0]
        registry_method = getattr(notifier_obj,
                                  '_send_nova_notification')
    elif notifier_method == DHCP_NOTIFIER_METHOD:
        # parse argument like "subnetpool.create.end"
        rname = args[2].split('.')[0]
        event_name = 'after_' + args[2].split('.')[1]
        registry_method = getattr(notifier_obj,
                                  '_native_event_send_dhcp_notification')
    if rname:
        cbacks = registry._get_callback_manager()._callbacks.get(rname, None)
        if cbacks and event_name in cbacks.keys():
            for entry in cbacks.values():
                method = entry.values()[0]
                if registry_method == method:
                    # This notification is already being sent by Neutron
                    # soe we will avoid sending a duplicate notification
                    return

    if not transaction_key or 'subnet.delete.end' in args or (
        not QUEUE_OUT_OF_PROCESS_NOTIFICATIONS):
        # We make an exception for the dhcp agent notification
        # for port and subnet delete since the implementation
        # for sending that notification checks for the existence
        # of the associated network, which is not present in certain
        # cases if the delete notification is queued and sent after
        # the network delete.
        getattr(notifier_obj, notifier_method)(*args)
        return

    _queue_notification(session, transaction_key, notifier_obj,
                        notifier_method, args)


def _get_callbacks_for_resource_event(resource, event):
    return list(registry._get_callback_manager()._callbacks[
        resource].get(event, {}).items())


def _get_in_process_callbacks(callbacks):
    return [i for i in callbacks if not [
        j for j in OUT_OF_PROCESS_NOTIFICATIONS if i[0].startswith(j)]]


def _registry_notify(resource, event, trigger, **kwargs):
    # This invokes neutron's original (unpatched) registry notification
    # method.
    registry._get_callback_manager().notify(
        resource, event, trigger, **kwargs)


def send_or_queue_registry_notification(
    session, transaction_key, resource, event, trigger, **kwargs):
    if not QUEUE_OUT_OF_PROCESS_NOTIFICATIONS:
        # Queueing is not enabled, so no more processing required,
        # relay notification to Neutron's callback registry
        _registry_notify(resource, event, trigger, **kwargs)
        return

    # Both, in-process and agent, notifieres may be registered for the
    # same event, so we might need to send and queue
    send, queue = False, False
    callbacks = _get_callbacks_for_resource_event(resource, event)
    if resource in ['port', 'router_interface', 'subnet'] and (
        event in ['after_update', 'after_delete', 'precommit_delete']):
        # We make an exception for the dhcp agent notification
        # for port and subnet since the implementation for
        # sending that notification checks for the existence of the
        # associated network, which is not present in certain
        # cases if the notification is queued and sent after the network
        # delete.
        # All notifiers (in-process as well as agent) will be
        # invoked in this case, no queueing of the notification
        # is required.
        send = True
    if not session or not transaction_key:
        # We can't queue notifications without session or transaction key
        send = True
    if not send:
        # Build a list of all in-process registered callbacks
        # for this resource
        in_process_callbacks = _get_in_process_callbacks(callbacks)
        # If there are notifiers registered which are not in-process,
        # we need to queue up this notification
        queue = (in_process_callbacks != callbacks)
        if in_process_callbacks:
            send = True
            # send only in-process callbacks
            callbacks = in_process_callbacks

    if send and callbacks:
        # Note: For the following to work, the _notify_loop()
        # function implemented in neutron.callbacks.manager
        # needs to be patched to handle the callbacks argument
        # like its being done in:
        # gbpservice/neutron/plugins/ml2plus/patch_neutron.py
        kwargs['callbacks'] = callbacks
        _registry_notify(resource, event, trigger, **kwargs)

    if queue and session and transaction_key:
        _queue_registry_notification(session, transaction_key, resource,
                                     event, trigger, **kwargs)


def post_notifications_from_queue(session, transaction_key):
    queue = session.notification_queue[transaction_key]
    for entry in queue:
        if REGISTRY_RESOURCE in entry:
            callbacks = _get_callbacks_for_resource_event(
                entry[REGISTRY_RESOURCE], entry[REGISTRY_EVENT])
            in_process_callbacks = _get_in_process_callbacks(callbacks)
            # Only process out-of-process notifications
            callbacks = list(set(callbacks) - set(in_process_callbacks))
            if callbacks:
                entry[NOTIFICATION_ARGS]['callbacks'] = callbacks
                _registry_notify(
                    entry[REGISTRY_RESOURCE], entry[REGISTRY_EVENT],
                    entry[REGISTRY_TRIGGER], **entry[NOTIFICATION_ARGS])
        else:
            getattr(entry[NOTIFIER_REF],
                    entry[NOTIFIER_METHOD])(*entry[NOTIFICATION_ARGS])
    del session.notification_queue[transaction_key]


def discard_notifications_after_rollback(session):
    session.notification_queue.pop(session.transaction, None)


class LocalAPI(object):
    """API for interacting with the neutron Plugins directly."""

    @property
    def _nova_notifier(self):
        return nova.Notifier()

    @property
    def _core_plugin(self):
        # REVISIT(rkukura): Need initialization method after all
        # plugins are loaded to grab and store plugin.
        return manager.NeutronManager.get_plugin()

    @property
    def _l3_plugin(self):
        # REVISIT(rkukura): Need initialization method after all
        # plugins are loaded to grab and store plugin.
        plugins = manager.NeutronManager.get_service_plugins()
        l3_plugin = plugins.get(pconst.L3_ROUTER_NAT)
        if not l3_plugin:
            LOG.error(_LE("No L3 router service plugin found."))
            raise exc.GroupPolicyDeploymentError()
        return l3_plugin

    @property
    def _group_policy_plugin(self):
        # REVISIT(rkukura): Need initialization method after all
        # plugins are loaded to grab and store plugin.
        plugins = manager.NeutronManager.get_service_plugins()
        group_policy_plugin = plugins.get(pconst.GROUP_POLICY)
        if not group_policy_plugin:
            LOG.error(_LE("No GroupPolicy service plugin found."))
            raise exc.GroupPolicyDeploymentError()
        return group_policy_plugin

    @property
    def _servicechain_plugin(self):
        # REVISIT(rkukura): Need initialization method after all
        # plugins are loaded to grab and store plugin.
        plugins = manager.NeutronManager.get_service_plugins()
        servicechain_plugin = plugins.get(pconst.SERVICECHAIN)
        if not servicechain_plugin:
            LOG.error(_LE("No Servicechain service plugin found."))
            raise exc.GroupPolicyDeploymentError()
        return servicechain_plugin

    @property
    def _trunk_plugin(self):
        # REVISIT(rkukura): Need initialization method after all
        # plugins are loaded to grab and store plugin.
        plugins = manager.NeutronManager.get_service_plugins()
        return plugins.get('trunk')

    def _create_resource(self, plugin, context, resource, attrs,
                         do_notify=True):
        # REVISIT(rkukura): Do create.start notification?
        # REVISIT(rkukura): Check authorization?
        reservation = None
        if plugin in [self._group_policy_plugin, self._servicechain_plugin]:
            reservation = quota.QUOTAS.make_reservation(
                context, context.tenant_id, {resource: 1}, plugin)
        action = 'create_' + resource
        obj_creator = getattr(plugin, action)
        try:
            obj = obj_creator(context, {resource: attrs})
        except Exception:
            # In case of failure the plugin will always raise an
            # exception. Cancel the reservation
            with excutils.save_and_reraise_exception():
                if reservation:
                    quota.QUOTAS.cancel_reservation(
                        context, reservation.reservation_id)
        if reservation:
            quota.QUOTAS.commit_reservation(
                context, reservation.reservation_id)
            # At this point the implicit resource creation is successfull,
            # so we should be calling:
            # resource_registry.set_resources_dirty(context)
            # to appropriately notify the quota engine. However, the above
            # call begins a new transaction and we want to avoid that.
            # Moreover, it can be safely assumed that any implicit resource
            # creation via this local_api is always in response to an
            # explicit resource creation request, and hence the above
            # method will be invoked in the API layer.
        return obj

    def _update_resource(self, plugin, context, resource, resource_id, attrs,
                         do_notify=True):
        # REVISIT(rkukura): Check authorization?
        action = 'update_' + resource
        obj_updater = getattr(plugin, action)
        obj = obj_updater(context, resource_id, {resource: attrs})
        return obj

    def _delete_resource(self, plugin, context, resource, resource_id,
                         do_notify=True):
        # REVISIT(rkukura): Check authorization?
        action = 'delete_' + resource
        obj_deleter = getattr(plugin, action)
        obj_deleter(context, resource_id)

    def _get_resource(self, plugin, context, resource, resource_id):
        obj_getter = getattr(plugin, 'get_' + resource)
        obj = obj_getter(context, resource_id)
        return obj

    def _get_resources(self, plugin, context, resource_plural, filters=None):
        obj_getter = getattr(plugin, 'get_' + resource_plural)
        obj = obj_getter(context, filters)
        return obj

    # The following methods perform the necessary subset of
    # functionality from neutron.api.v2.base.Controller.
    #
    # REVISIT(rkukura): Can we just use the WSGI Controller?  Using
    # neutronclient is also a possibility, but presents significant
    # issues to unit testing as well as overhead and failure modes.

    def _get_port(self, plugin_context, port_id):
        return self._get_resource(self._core_plugin, plugin_context, 'port',
                                  port_id)

    def _get_ports(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._core_plugin, plugin_context, 'ports',
                                   filters)

    def _create_port(self, plugin_context, attrs):
        return self._create_resource(self._core_plugin, plugin_context, 'port',
                                     attrs)

    def _update_port(self, plugin_context, port_id, attrs):
        return self._update_resource(self._core_plugin, plugin_context, 'port',
                                     port_id, attrs)

    def _delete_port(self, plugin_context, port_id):
        try:
            self._delete_resource(self._core_plugin,
                                  plugin_context, 'port', port_id)
        except n_exc.PortNotFound:
            LOG.warning(_LW('Port %s already deleted'), port_id)

    def _get_subnet(self, plugin_context, subnet_id):
        return self._get_resource(self._core_plugin, plugin_context, 'subnet',
                                  subnet_id)

    def _get_subnets(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._core_plugin, plugin_context,
                                   'subnets', filters)

    def _create_subnet(self, plugin_context, attrs):
        return self._create_resource(self._core_plugin, plugin_context,
                                     'subnet', attrs)

    def _update_subnet(self, plugin_context, subnet_id, attrs):
        return self._update_resource(self._core_plugin, plugin_context,
                                     'subnet', subnet_id, attrs)

    def _delete_subnet(self, plugin_context, subnet_id):
        try:
            self._delete_resource(self._core_plugin, plugin_context, 'subnet',
                                  subnet_id)
        except n_exc.SubnetNotFound:
            LOG.warning(_LW('Subnet %s already deleted'), subnet_id)

    def _get_network(self, plugin_context, network_id):
        return self._get_resource(self._core_plugin, plugin_context, 'network',
                                  network_id)

    def _get_networks(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(
            self._core_plugin, plugin_context, 'networks', filters)

    def _create_network(self, plugin_context, attrs):
        return self._create_resource(self._core_plugin, plugin_context,
                                     'network', attrs, True)

    def _update_network(self, plugin_context, network_id, attrs):
        return self._update_resource(self._core_plugin, plugin_context,
                                     'network', network_id, attrs)

    def _delete_network(self, plugin_context, network_id):
        try:
            self._delete_resource(self._core_plugin, plugin_context,
                                  'network', network_id)
        except n_exc.NetworkNotFound:
            LOG.warning(_LW('Network %s already deleted'), network_id)

    def _get_router(self, plugin_context, router_id):
        return self._get_resource(self._l3_plugin, plugin_context, 'router',
                                  router_id)

    def _get_routers(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._l3_plugin, plugin_context, 'routers',
                                   filters)

    def _create_router(self, plugin_context, attrs):
        return self._create_resource(self._l3_plugin, plugin_context, 'router',
                                     attrs)

    def _update_router(self, plugin_context, router_id, attrs):
        return self._update_resource(self._l3_plugin, plugin_context, 'router',
                                     router_id, attrs)

    def _add_router_interface(self, plugin_context, router_id, interface_info):
        self._l3_plugin.add_router_interface(plugin_context,
                                             router_id, interface_info)

    def _remove_router_interface(self, plugin_context, router_id,
                                 interface_info):
        # To detach Router interface either port ID or Subnet ID is mandatory
        try:
            self._l3_plugin.remove_router_interface(plugin_context, router_id,
                                                    interface_info)
        except l3.RouterInterfaceNotFoundForSubnet:
            LOG.warning(_LW('Router interface already deleted for subnet %s'),
                        interface_info)
            return

    def _add_router_gw_interface(self, plugin_context, router_id, gw_info):
        return self._l3_plugin.update_router(
            plugin_context, router_id,
            {'router': {'external_gateway_info': gw_info}})

    def _remove_router_gw_interface(self, plugin_context, router_id,
                                    interface_info):
        self._l3_plugin.update_router(
            plugin_context, router_id,
            {'router': {'external_gateway_info': None}})

    def _delete_router(self, plugin_context, router_id):
        try:
            self._delete_resource(self._l3_plugin, plugin_context, 'router',
                                  router_id)
        except l3.RouterNotFound:
            LOG.warning(_LW('Router %s already deleted'), router_id)

    def _get_sg(self, plugin_context, sg_id):
        return self._get_resource(
            self._core_plugin, plugin_context, 'security_group', sg_id)

    def _get_sgs(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(
            self._core_plugin, plugin_context, 'security_groups', filters)

    def _create_sg(self, plugin_context, attrs):
        return self._create_resource(self._core_plugin, plugin_context,
                                     'security_group', attrs)

    def _update_sg(self, plugin_context, sg_id, attrs):
        return self._update_resource(self._core_plugin, plugin_context,
                                     'security_group', sg_id, attrs)

    def _delete_sg(self, plugin_context, sg_id):
        try:
            self._delete_resource(self._core_plugin, plugin_context,
                                  'security_group', sg_id)
        except ext_sg.SecurityGroupNotFound:
            LOG.warning(_LW('Security Group %s already deleted'), sg_id)

    def _get_sg_rule(self, plugin_context, sg_rule_id):
        return self._get_resource(
            self._core_plugin, plugin_context, 'security_group_rule',
            sg_rule_id)

    def _get_sg_rules(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(
            self._core_plugin, plugin_context, 'security_group_rules', filters)

    def _create_sg_rule(self, plugin_context, attrs):
        try:
            return self._create_resource(self._core_plugin, plugin_context,
                                         'security_group_rule', attrs)
        except ext_sg.SecurityGroupRuleExists as ex:
            LOG.warning(_LW('Security Group already exists %s'), ex.message)
            return

    def _update_sg_rule(self, plugin_context, sg_rule_id, attrs):
        return self._update_resource(self._core_plugin, plugin_context,
                                     'security_group_rule', sg_rule_id,
                                     attrs)

    def _delete_sg_rule(self, plugin_context, sg_rule_id):
        try:
            self._delete_resource(self._core_plugin, plugin_context,
                                  'security_group_rule', sg_rule_id)
        except ext_sg.SecurityGroupRuleNotFound:
            LOG.warning(_LW('Security Group Rule %s already deleted'),
                        sg_rule_id)

    def _get_fip(self, plugin_context, fip_id):
        return self._get_resource(
            self._l3_plugin, plugin_context, 'floatingip', fip_id)

    def _get_fips(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(
            self._l3_plugin, plugin_context, 'floatingips', filters)

    def _create_fip(self, plugin_context, attrs):
        return self._create_resource(self._l3_plugin, plugin_context,
                                     'floatingip', attrs)

    def _update_fip(self, plugin_context, fip_id, attrs):
        return self._update_resource(self._l3_plugin, plugin_context,
                                     'floatingip', fip_id, attrs)

    def _delete_fip(self, plugin_context, fip_id):
        try:
            self._delete_resource(self._l3_plugin, plugin_context,
                                  'floatingip', fip_id)
        except l3.FloatingIPNotFound:
            LOG.warning(_LW('Floating IP %s Already deleted'), fip_id)

    def _get_address_scope(self, plugin_context, address_scope_id):
        return self._get_resource(self._core_plugin, plugin_context,
                                  'address_scope', address_scope_id)

    def _get_address_scopes(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._core_plugin, plugin_context,
                                   'address_scopes', filters)

    def _create_address_scope(self, plugin_context, attrs):
        return self._create_resource(self._core_plugin, plugin_context,
                                     'address_scope', attrs)

    def _update_address_scope(self, plugin_context, address_scope_id, attrs):
        return self._update_resource(self._core_plugin, plugin_context,
                                     'address_scope', address_scope_id, attrs)

    def _delete_address_scope(self, plugin_context, address_scope_id):
        try:
            self._delete_resource(self._core_plugin, plugin_context,
                                  'address_scope', address_scope_id)
        except address_scope.AddressScopeNotFound:
            LOG.warning(_LW('Address Scope %s already deleted'),
                        address_scope_id)

    def _get_subnetpool(self, plugin_context, subnetpool_id):
        return self._get_resource(self._core_plugin, plugin_context,
                                  'subnetpool', subnetpool_id)

    def _get_subnetpools(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._core_plugin, plugin_context,
                                   'subnetpools', filters)

    def _create_subnetpool(self, plugin_context, attrs):
        return self._create_resource(self._core_plugin, plugin_context,
                                     'subnetpool', attrs)

    def _update_subnetpool(self, plugin_context, subnetpool_id, attrs):
        return self._update_resource(self._core_plugin, plugin_context,
                                     'subnetpool', subnetpool_id, attrs)

    def _delete_subnetpool(self, plugin_context, subnetpool_id):
        try:
            self._delete_resource(self._core_plugin, plugin_context,
                                  'subnetpool', subnetpool_id)
        except n_exc.SubnetpoolNotFound:
            LOG.warning(_LW('Subnetpool %s already deleted'), subnetpool_id)

    def _get_l2_policy(self, plugin_context, l2p_id):
        return self._get_resource(self._group_policy_plugin, plugin_context,
                                  'l2_policy', l2p_id)

    def _get_l2_policies(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._group_policy_plugin, plugin_context,
                                   'l2_policies', filters)

    def _create_l2_policy(self, plugin_context, attrs):
        return self._create_resource(self._group_policy_plugin, plugin_context,
                                     'l2_policy', attrs, False)

    def _update_l2_policy(self, plugin_context, l2p_id, attrs):
        return self._update_resource(self._group_policy_plugin, plugin_context,
                                     'l2_policy', l2p_id, attrs, False)

    def _delete_l2_policy(self, plugin_context, l2p_id):
        try:
            self._delete_resource(self._group_policy_plugin,
                                  plugin_context, 'l2_policy', l2p_id, False)
        except gp_ext.L2PolicyNotFound:
            LOG.warning(_LW('L2 Policy %s already deleted'), l2p_id)

    def _get_l3_policy(self, plugin_context, l3p_id):
        return self._get_resource(self._group_policy_plugin, plugin_context,
                                  'l3_policy', l3p_id)

    def _get_l3_policies(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._group_policy_plugin, plugin_context,
                                   'l3_policies', filters)

    def _create_l3_policy(self, plugin_context, attrs):
        return self._create_resource(self._group_policy_plugin, plugin_context,
                                     'l3_policy', attrs, False)

    def _update_l3_policy(self, plugin_context, l3p_id, attrs):
        return self._update_resource(self._group_policy_plugin, plugin_context,
                                     'l3_policy', l3p_id, attrs, False)

    def _delete_l3_policy(self, plugin_context, l3p_id):
        try:
            self._delete_resource(self._group_policy_plugin,
                                  plugin_context, 'l3_policy', l3p_id, False)
        except gp_ext.L3PolicyNotFound:
            LOG.warning(_LW('L3 Policy %s already deleted'), l3p_id)

    def _get_external_segment(self, plugin_context, es_id):
        return self._get_resource(self._group_policy_plugin, plugin_context,
                                  'external_segment', es_id)

    def _get_external_segments(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._group_policy_plugin, plugin_context,
                                   'external_segments', filters)

    def _create_external_segment(self, plugin_context, attrs):
        return self._create_resource(self._group_policy_plugin, plugin_context,
                                     'external_segment', attrs, False)

    def _update_external_segment(self, plugin_context, es_id, attrs):
        return self._update_resource(self._group_policy_plugin, plugin_context,
                                     'external_segment', es_id, attrs, False)

    def _delete_external_segment(self, plugin_context, es_id):
        try:
            self._delete_resource(self._group_policy_plugin, plugin_context,
                                  'external_segment', es_id, False)
        except gp_ext.ExternalSegmentNotFound:
            LOG.warning(_LW('External Segment %s already deleted'), es_id)

    def _get_external_policy(self, plugin_context, ep_id):
        return self._get_resource(self._group_policy_plugin, plugin_context,
                                  'external_policy', ep_id)

    def _get_external_policies(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._group_policy_plugin, plugin_context,
                                   'external_policies', filters)

    def _create_external_policy(self, plugin_context, attrs):
        return self._create_resource(self._group_policy_plugin, plugin_context,
                                     'external_policy', attrs, False)

    def _update_external_policy(self, plugin_context, ep_id, attrs):
        return self._update_resource(self._group_policy_plugin, plugin_context,
                                     'external_policy', ep_id, attrs, False)

    def _delete_external_policy(self, plugin_context, ep_id):
        try:
            self._delete_resource(self._group_policy_plugin, plugin_context,
                                  'external_policy', ep_id, False)
        except gp_ext.ExternalPolicyNotFound:
            LOG.warning(_LW('External Policy %s already deleted'), ep_id)

    def _get_policy_rule_set(self, plugin_context, prs_id):
        return self._get_resource(self._group_policy_plugin, plugin_context,
                                  'policy_rule_set', prs_id)

    def _get_policy_rule_sets(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._group_policy_plugin, plugin_context,
                                   'policy_rule_sets', filters)

    def _create_policy_rule_set(self, plugin_context, attrs):
        return self._create_resource(self._group_policy_plugin, plugin_context,
                                     'policy_rule_set', attrs, False)

    def _update_policy_rule_set(self, plugin_context, prs_id, attrs):
        return self._update_resource(self._group_policy_plugin, plugin_context,
                                     'policy_rule_set', prs_id, attrs, False)

    def _delete_policy_rule_set(self, plugin_context, prs_id):
        try:
            self._delete_resource(self._group_policy_plugin, plugin_context,
                                  'policy_rule_set', prs_id, False)
        except gp_ext.PolicyRuleSetNotFound:
            LOG.warning(_LW('Policy Rule Set %s already deleted'), prs_id)

    def _get_servicechain_instance(self, plugin_context, sci_id):
        return self._get_resource(self._servicechain_plugin, plugin_context,
                                  'servicechain_instance', sci_id)

    def _get_servicechain_instances(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._servicechain_plugin, plugin_context,
                                   'servicechain_instances', filters)

    def _create_servicechain_instance(self, plugin_context, attrs):
        return self._create_resource(self._servicechain_plugin, plugin_context,
                                     'servicechain_instance', attrs, False)

    def _update_servicechain_instance(self, plugin_context, sci_id, attrs):
        return self._update_resource(self._servicechain_plugin, plugin_context,
                                     'servicechain_instance', sci_id, attrs,
                                     False)

    def _delete_servicechain_instance(self, plugin_context, sci_id):
        try:
            self._delete_resource(self._servicechain_plugin, plugin_context,
                                  'servicechain_instance', sci_id, False)
        except sc_ext.ServiceChainInstanceNotFound:
            LOG.warning(_LW("servicechain %s already deleted"), sci_id)

    def _get_servicechain_spec(self, plugin_context, scs_id):
        return self._get_resource(self._servicechain_plugin, plugin_context,
                                  'servicechain_spec', scs_id)

    def _get_servicechain_specs(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._servicechain_plugin, plugin_context,
                                   'servicechain_specs', filters)

    def _create_servicechain_spec(self, plugin_context, attrs):
        return self._create_resource(self._servicechain_plugin, plugin_context,
                                     'servicechain_spec', attrs, False)

    def _update_servicechain_spec(self, plugin_context, scs_id, attrs):
        return self._update_resource(self._servicechain_plugin, plugin_context,
                                     'servicechain_spec', scs_id, attrs, False)

    def _delete_servicechain_spec(self, plugin_context, scs_id):
        try:
            self._delete_resource(self._servicechain_plugin, plugin_context,
                                  'servicechain_spec', scs_id)
        except sc_ext.ServiceChainSpecNotFound:
            LOG.warning(_LW("servicechain spec %s already deleted"), scs_id)

    def _get_policy_target(self, plugin_context, pt_id):
        return self._get_resource(self._group_policy_plugin, plugin_context,
                                  'policy_target', pt_id)

    def _get_policy_targets(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._group_policy_plugin, plugin_context,
                                   'policy_targets', filters)

    def _create_policy_target(self, plugin_context, attrs):
        return self._create_resource(self._group_policy_plugin, plugin_context,
                                     'policy_target', attrs, False)

    def _update_policy_target(self, plugin_context, pt_id, attrs):
        return self._update_resource(self._group_policy_plugin, plugin_context,
                                     'policy_target', pt_id, attrs, False)

    def _delete_policy_target(self, plugin_context, pt_id):
        try:
            self._delete_resource(self._group_policy_plugin, plugin_context,
                                  'policy_target', pt_id, False)
        except gp_ext.PolicyTargetNotFound:
            LOG.warning(_LW('Policy Rule Set %s already deleted'), pt_id)

    def _get_policy_target_group(self, plugin_context, ptg_id):
        return self._get_resource(self._group_policy_plugin, plugin_context,
                                  'policy_target_group', ptg_id)

    def _get_policy_target_groups(self, plugin_context, filters=None):
        filters = filters or {}
        return self._get_resources(self._group_policy_plugin, plugin_context,
                                   'policy_target_groups', filters)

    def _create_policy_target_group(self, plugin_context, attrs):
        return self._create_resource(self._group_policy_plugin, plugin_context,
                                     'policy_target_group', attrs, False)

    def _update_policy_target_group(self, plugin_context, ptg_id, attrs):
        return self._update_resource(self._group_policy_plugin, plugin_context,
                                     'policy_target_group', ptg_id, attrs,
                                     False)

    def _delete_policy_target_group(self, plugin_context, ptg_id):
        try:
            self._delete_resource(self._group_policy_plugin, plugin_context,
                                  'policy_target_group', ptg_id)
        except sc_ext.ServiceChainSpecNotFound:
            LOG.warning(_LW("Policy Target Group %s already deleted"), ptg_id)



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""servicechain

Revision ID: ebfd08bc4714
Revises: 53de98f7a066
Create Date: 2014-07-24 16:12:22.610815

"""

# revision identifiers, used by Alembic.
revision = 'ebfd08bc4714'
down_revision = '5c65abe72596'


from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.create_table(
        'sc_nodes',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=50), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('service_type', sa.String(length=50), nullable=True),
        sa.Column('config', sa.String(length=4096), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'sc_specs',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=50), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('config_param_names', sa.String(length=4096), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'sc_instances',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=50), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.Column('config_param_values', sa.String(length=4096),
                  nullable=True),
        sa.Column('provider_ptg_id', sa.String(length=36), nullable=True),
        sa.Column('consumer_ptg_id', sa.String(length=36), nullable=True),
        sa.Column('classifier_id', sa.String(length=36), nullable=True),
        # FixMe(Magesh) If cascade on delete is used, we lose this info !!!
        # sa.ForeignKeyConstraint(['provider_ptg_id'],
        #                         ['gp_policy_target_groups.id'],
        #                         ondelete='CASCADE'),
        # sa.ForeignKeyConstraint(['consumer_ptg_id'],
        #                         ['gp_policy_target_groups.id'],
        #                         ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'sc_spec_node_associations',
        sa.Column('servicechain_spec_id',
                  sa.String(length=36),
                  nullable=False),
        sa.Column('node_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['servicechain_spec_id'], ['sc_specs.id']),
        sa.ForeignKeyConstraint(['node_id'], ['sc_nodes.id']),
        sa.PrimaryKeyConstraint('servicechain_spec_id', 'node_id')
    )


def downgrade(active_plugins=None, options=None):
    op.drop_table('sc_spec_node_associations')
    op.drop_table('sc_instances')
    op.drop_table('sc_specs')
    op.drop_table('sc_nodes')



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import copy
import six

from neutron.plugins.common import constants
from neutron_lib import constants as n_constants

from gbpservice.neutron.extensions import group_policy as gp
from gbpservice.neutron.extensions import group_policy_mapping as gpm
from gbpservice.neutron.tests.unit import common as cm
from gbpservice.neutron.tests.unit import test_extension_group_policy as tgp


class GroupPolicyMappingExtTestCase(tgp.GroupPolicyExtensionTestCase):
    def setUp(self):
        self._saved_gp_attr_map = {}
        for k, v in six.iteritems(gp.RESOURCE_ATTRIBUTE_MAP):
            self._saved_gp_attr_map[k] = v.copy()
        self.addCleanup(self._restore_gp_attr_map)

        super(tgp.GroupPolicyExtensionTestCase, self).setUp()

        attr_map = gp.RESOURCE_ATTRIBUTE_MAP
        attr_map[gp.POLICY_TARGETS].update(
            gpm.EXTENDED_ATTRIBUTES_2_0[gp.POLICY_TARGETS])
        attr_map[gp.POLICY_TARGET_GROUPS].update(
            gpm.EXTENDED_ATTRIBUTES_2_0[gp.POLICY_TARGET_GROUPS])
        attr_map[gp.L2_POLICIES].update(
            gpm.EXTENDED_ATTRIBUTES_2_0[gp.L2_POLICIES])
        attr_map[gp.L3_POLICIES].update(
            gpm.EXTENDED_ATTRIBUTES_2_0[gp.L3_POLICIES])
        attr_map[gp.EXTERNAL_SEGMENTS].update(
            gpm.EXTENDED_ATTRIBUTES_2_0[gp.EXTERNAL_SEGMENTS])
        plural_mappings = {'l2_policy': 'l2_policies',
                           'l3_policy': 'l3_policies',
                           'network_service_policy':
                           'network_service_policies',
                           'external_policy':
                           'external_policies'}
        self._setUpExtension(
            tgp.GP_PLUGIN_BASE_NAME, constants.GROUP_POLICY, attr_map,
            gp.Group_policy, tgp.GROUPPOLICY_URI,
            plural_mappings=plural_mappings)
        self.instance = self.plugin.return_value

    def _restore_gp_attr_map(self):
        gp.RESOURCE_ATTRIBUTE_MAP = self._saved_gp_attr_map

    def get_create_policy_target_default_attrs(self):
        attrs = cm.get_create_policy_target_default_attrs()
        attrs.update({'port_id': None})
        return attrs

    def get_create_policy_target_default_attrs_and_prj_id(self):
        attrs = cm.get_create_policy_target_default_attrs_and_prj_id()
        attrs.update({'port_id': None})
        return attrs

    def get_create_policy_target_attrs(self):
        attrs = cm.get_create_policy_target_attrs()
        attrs.update({'port_id': tgp._uuid()})
        fixed_ips = [{'subnet_id': '00000000-ffff-ffff-ffff-000000000000',
                      'ip_address': '11.1.1.1'}]
        attrs.update({'fixed_ips': fixed_ips})
        return attrs

    def get_update_policy_target_attrs(self):
        attrs = cm.get_update_policy_target_attrs()
        fixed_ips = [{'subnet_id': '00000000-ffff-ffff-ffff-000000000000',
                      'ip_address': '11.1.1.1'}]
        attrs.update({'fixed_ips': fixed_ips})
        return attrs

    def get_create_policy_target_group_default_attrs(self):
        attrs = cm.get_create_policy_target_group_default_attrs()
        attrs.update({'subnets': []})
        return attrs

    def get_create_policy_target_group_default_attrs_and_prj_id(self):
        attrs = cm.get_create_policy_target_group_default_attrs_and_prj_id()
        attrs.update({'subnets': []})
        return attrs

    def get_create_policy_target_group_attrs(self):
        attrs = cm.get_create_policy_target_group_attrs()
        attrs.update({'subnets': [tgp._uuid()]})
        return attrs

    def get_update_policy_target_group_attrs(self):
        attrs = cm.get_update_policy_target_group_attrs()
        attrs.update({'subnets': [tgp._uuid()]})
        return attrs

    def get_create_l2_policy_default_attrs(self):
        attrs = cm.get_create_l2_policy_default_attrs()
        attrs.update({'network_id': None})
        return attrs

    def get_create_l2_policy_default_attrs_and_prj_id(self):
        attrs = cm.get_create_l2_policy_default_attrs_and_prj_id()
        attrs.update({'network_id': None})
        return attrs

    def get_create_l2_policy_attrs(self):
        attrs = cm.get_create_l2_policy_attrs()
        attrs.update({'network_id': tgp._uuid()})
        return attrs

    def get_create_l3_policy_default_attrs(self):
        attrs = cm.get_create_l3_policy_default_attrs()
        attrs.update({'address_scope_v4_id': None})
        attrs.update({'address_scope_v6_id': None})
        attrs.update({'subnetpools_v4': []})
        attrs.update({'subnetpools_v6': []})
        attrs.update({'routers': []})
        return attrs

    def get_create_l3_policy_default_attrs_and_prj_id(self):
        attrs = cm.get_create_l3_policy_default_attrs_and_prj_id()
        attrs.update({'address_scope_v4_id': None})
        attrs.update({'address_scope_v6_id': None})
        attrs.update({'subnetpools_v4': []})
        attrs.update({'subnetpools_v6': []})
        attrs.update({'routers': []})
        return attrs

    def get_create_l3_policy_attrs(self):
        attrs = cm.get_create_l3_policy_attrs()
        attrs.update({'address_scope_v4_id': tgp._uuid()})
        attrs.update({'address_scope_v6_id': tgp._uuid()})
        attrs.update({'subnetpools_v4': [tgp._uuid(), tgp._uuid()]})
        attrs.update({'subnetpools_v6': [tgp._uuid(), tgp._uuid()]})
        attrs.update({'routers': [tgp._uuid(), tgp._uuid()]})
        return attrs

    def get_update_l3_policy_attrs(self):
        attrs = cm.get_update_l3_policy_attrs()
        attrs.update({'subnetpools_v4': [tgp._uuid(), tgp._uuid()]})
        attrs.update({'subnetpools_v6': [tgp._uuid(), tgp._uuid()]})
        attrs.update({'routers': [tgp._uuid(), tgp._uuid()]})
        return attrs

    def get_create_external_segment_default_attrs(self):
        attrs = cm.get_create_external_segment_default_attrs()
        attrs.update({'subnet_id': None})
        return attrs

    def get_create_external_segment_default_attrs_and_prj_id(self):
        attrs = cm.get_create_external_segment_default_attrs_and_prj_id()
        attrs.update({'subnet_id': None})
        return attrs

    def get_create_external_segment_attrs(self):
        attrs = cm.get_create_external_segment_attrs()
        attrs.update({'subnet_id': tgp._uuid()})
        return attrs

    def test_create_policy_target_with_defaults(self):
        policy_target_id = tgp._uuid()
        data = {'policy_target': {'policy_target_group_id': tgp._uuid(),
                                  'tenant_id': tgp._uuid()}}
        default_attrs = (
            self.get_create_policy_target_default_attrs_and_prj_id())
        default_data = copy.copy(data)
        default_data['policy_target'].update(default_attrs)
        expected_value = dict(default_data['policy_target'])
        expected_value['id'] = policy_target_id
        expected_value['fixed_ips'] = n_constants.ATTR_NOT_SPECIFIED

        self._test_create_policy_target(data, expected_value, default_data)
